import json
import uuid
from typing import List, Dict, Any, Optional
from ..interfaces import VectorStore
from ..config import DatabaseConfig

class PrismaVectorStore(VectorStore):
    def __init__(self, config: DatabaseConfig):
        self.config = config

    def _safe_ident(self, value: str) -> str:
        import re
        v = str(value or "")
        if not re.fullmatch(r"[A-Za-z_][A-Za-z0-9_]*", v):
            raise ValueError(f"Unsafe SQL identifier: {v!r}")
        return v
    
    async def ensure_indexes(self):
        client = self.config.client_instance
        table = self._safe_ident(self.config.table_name or "Document")
        c_vec = self._safe_ident(self.config.column_map.get('vector', 'embedding'))
        c_content = self._safe_ident(self.config.column_map.get('content', 'content'))
        try:
            await client.execute_raw("CREATE EXTENSION IF NOT EXISTS vector")
            await client.execute_raw(f"""
                CREATE INDEX IF NOT EXISTS "{table}_embedding_ivfflat"
                ON "{table}" USING ivfflat ("{c_vec}" vector_cosine_ops)
                WITH (lists = 100);
            """)
            await client.execute_raw(f"""
                CREATE INDEX IF NOT EXISTS "{table}_content_fts_gin"
                ON "{table}" USING GIN (to_tsvector('english', "{c_content}"));
            """)
        except Exception:
            pass

    def normalize_vector(self, v: List[float]) -> List[float]:
        norm = sum(x * x for x in v) ** 0.5
        if norm == 0: return v
        return [x / norm for x in v]

    async def add_documents(self, documents: List[Dict[str, Any]]):
        client = self.config.client_instance
        table = self._safe_ident(self.config.table_name or "Document")
        # Assuming flexible column mapping, defaulting to standard names
        col_content = self._safe_ident(self.config.column_map.get('content', 'content'))
        col_meta = self._safe_ident(self.config.column_map.get('metadata', 'metadata'))
        col_vec = self._safe_ident(self.config.column_map.get('vector', 'embedding'))

        for doc in documents:
            vec_str = json.dumps(self.normalize_vector(doc['embedding']))
            query = f"""
            INSERT INTO "{table}" ("id", "{col_content}", "{col_meta}", "{col_vec}", "createdAt")
            VALUES ($1::uuid, $2, $3::jsonb, $4::vector, NOW())
            """
            await client.execute_raw(
                query,
                str(uuid.uuid4()),
                doc['content'],
                json.dumps(doc['metadata']),
                vec_str
            )
    
    async def file_exists(self, sha256: str, size: int, last_modified: int) -> bool:
        client = self.config.client_instance
        table = self._safe_ident(self.config.table_name or "Document")
        col_meta = self._safe_ident(self.config.column_map.get('metadata', 'metadata'))
        payload = json.dumps({"fileSHA256": sha256, "fileSize": size, "lastModified": last_modified})
        q = f'SELECT 1 FROM "{table}" WHERE "{col_meta}" @> $1::jsonb LIMIT 1'
        try:
            res = await client.query_raw(q, payload)
            return len(res) > 0
        except Exception:
            return False

    async def similarity_search(self, vector: List[float], limit: int = 5, filter: Optional[Dict] = None) -> List[Dict[str, Any]]:
        client = self.config.client_instance
        table = self._safe_ident(self.config.table_name or "Document")
        col_content = self._safe_ident(self.config.column_map.get('content', 'content'))
        col_meta = self._safe_ident(self.config.column_map.get('metadata', 'metadata'))
        col_vec = self._safe_ident(self.config.column_map.get('vector', 'embedding'))
        
        vec_str = json.dumps(self.normalize_vector(vector))
        where_clause = ""
        params = [vec_str]
        
        if filter:
            where_clause = f'WHERE "{col_meta}" @> $2::jsonb'
            params.append(json.dumps(filter))
            
        query = f"""
        SELECT "{col_content}" as content, "{col_meta}" as metadata, 1 - ("{col_vec}" <=> $1::vector) as score
        FROM "{table}"
        {where_clause}
        ORDER BY score DESC
        LIMIT {limit}
        """
        
        if len(params) == 1: res = await client.query_raw(query, params[0])
        else: res = await client.query_raw(query, params[0], params[1])
            
        return [
            {
                'content': r['content'], 
                'metadata': r['metadata'] if isinstance(r['metadata'], dict) else json.loads(r['metadata']), 
                'score': float(r['score'])
            } for r in res
        ]
    
    async def hybrid_search(self, text: str, vector: List[float], limit: int = 5, filter: Optional[Dict] = None) -> List[Dict[str, Any]]:
        semantic = await self.similarity_search(vector, limit * 2, filter)
        client = self.config.client_instance
        table = self._safe_ident(self.config.table_name or "Document")
        c_content = self._safe_ident(self.config.column_map.get('content', 'content'))
        c_metadata = self._safe_ident(self.config.column_map.get('metadata', 'metadata'))
        where = f"to_tsvector('simple', \"{c_content}\") @@ plainto_tsquery($1)"
        params: List[Any] = [text]
        if filter:
            where = where + f' AND "{c_metadata}" @> $2::jsonb'
            params.append(json.dumps(filter))
        q = f"SELECT \"{c_content}\" as content, \"{c_metadata}\" as metadata FROM \"{table}\" WHERE {where} LIMIT {limit * 2}"
        lexical = []
        try:
            res = await client.query_raw(q, *params)
            lexical = [{ 'content': r['content'], 'metadata': r.get('metadata'), 'score': 1.0 } for r in res]
        except Exception:
            lexical = []
        combined: Dict[str, Dict[str, Any]] = {}
        def add(lst: List[Dict[str, Any]], weight: float = 1.0):
            for idx, doc in enumerate(lst):
                key = doc['content']
                score = (1.0 / (60 + idx + 1)) * weight
                if key not in combined:
                    combined[key] = { **doc, 'score': 0.0 }
                combined[key]['score'] += score
        add(semantic, 1.0)
        add(lexical, 1.0)
        return sorted(combined.values(), key=lambda d: d['score'], reverse=True)[:limit]

    async def list_documents(self, filter: Optional[Dict[str, Any]] = None, limit: int = 100) -> List[Dict[str, Any]]:
        client = self.config.client_instance
        table = self._safe_ident(self.config.table_name or "Document")
        c_content = self._safe_ident(self.config.column_map.get("content", "content"))
        c_metadata = self._safe_ident(self.config.column_map.get("metadata", "metadata"))
        limit_int = max(1, int(limit))
        where = ""
        params: List[Any] = []
        if filter:
            where = f'WHERE "{c_metadata}" @> $1::jsonb'
            params.append(json.dumps(filter))
        q = f"""
        SELECT "id" as id, "{c_content}" as content, "{c_metadata}" as metadata
        FROM "{table}"
        {where}
        ORDER BY "createdAt" DESC
        LIMIT {limit_int}
        """
        if params:
            rows = await client.query_raw(q, params[0])
        else:
            rows = await client.query_raw(q)
        out: List[Dict[str, Any]] = []
        for r in rows:
            md = r.get("metadata")
            out.append(
                {
                    "id": r.get("id"),
                    "content": r.get("content"),
                    "metadata": md if isinstance(md, dict) else json.loads(md) if isinstance(md, str) else md,
                }
            )
        return out

    async def delete_documents(self, filter: Dict[str, Any]) -> int:
        client = self.config.client_instance
        table = self._safe_ident(self.config.table_name or "Document")
        c_metadata = self._safe_ident(self.config.column_map.get("metadata", "metadata"))
        q = f'DELETE FROM "{table}" WHERE "{c_metadata}" @> $1::jsonb RETURNING 1'
        rows = await client.query_raw(q, json.dumps(filter))
        return len(rows)

    async def update_documents(self, filter: Dict[str, Any], update_data: Dict[str, Any]) -> int:
        client = self.config.client_instance
        table = self._safe_ident(self.config.table_name or "Document")
        c_content = self._safe_ident(self.config.column_map.get("content", "content"))
        c_metadata = self._safe_ident(self.config.column_map.get("metadata", "metadata"))

        sets: List[str] = []
        params: List[Any] = []
        idx = 1

        if "content" in update_data and update_data["content"] is not None:
            sets.append(f'"{c_content}" = ${idx}')
            params.append(str(update_data["content"]))
            idx += 1

        if "metadata" in update_data and isinstance(update_data["metadata"], dict):
            sets.append(f'"{c_metadata}" = COALESCE("{c_metadata}", \'{{}}\'::jsonb) || ${idx}::jsonb')
            params.append(json.dumps(update_data["metadata"]))
            idx += 1

        if not sets:
            return 0

        where = f'"{c_metadata}" @> ${idx}::jsonb'
        params.append(json.dumps(filter))

        q = f'UPDATE "{table}" SET {", ".join(sets)} WHERE {where} RETURNING 1'
        rows = await client.query_raw(q, *params)
        return len(rows)



import os
import asyncio
from typing import List, AsyncGenerator
import json
import urllib.request
from ..config import LLMConfig

class HuggingFaceBackend:
    def __init__(self, config: LLMConfig):
        self.config = config
        self.api_key = config.api_key or os.getenv('HUGGINGFACE_API_KEY')
        if not self.api_key:
            raise ValueError('HuggingFace API Key missing. Set HUGGINGFACE_API_KEY.')
        self.base_url = 'https://api-inference.huggingface.co/models'

    async def _post(self, model: str, payload: dict):
        def _do():
            req = urllib.request.Request(
                url=f"{self.base_url}/{model}",
                data=json.dumps(payload).encode('utf-8'),
                headers={ 'Authorization': f'Bearer {self.api_key}', 'Content-Type': 'application/json' },
                method='POST'
            )
            with urllib.request.urlopen(req) as resp:
                return json.loads(resp.read().decode('utf-8'))
        return await asyncio.to_thread(_do)

    async def embed_documents(self, texts: List[str]) -> List[List[float]]:
        out: List[List[float]] = []
        for text in texts:
            r = await self._post(self.config.model_name, { 'inputs': text, 'options': { 'wait_for_model': True } })
            vec = r if isinstance(r, list) else r.get('embedding')
            flat = []
            def _flatten(x):
                if isinstance(x, list):
                    for i in x: _flatten(i)
                else:
                    try: flat.append(float(x))
                    except: pass
            _flatten(vec)
            out.append(flat)
        return out

    async def embed_query(self, text: str) -> List[float]:
        r = await self._post(self.config.model_name, { 'inputs': text, 'options': { 'wait_for_model': True } })
        vec = r if isinstance(r, list) else r.get('embedding')
        flat: List[float] = []
        def _flatten(x):
            if isinstance(x, list):
                for i in x: _flatten(i)
            else:
                try: flat.append(float(x))
                except: pass
        _flatten(vec)
        return flat

    async def generate(self, prompt: str, system_instruction: str = '') -> str:
        inputs = f"{system_instruction}\n{prompt}" if system_instruction else prompt
        r = await self._post(self.config.model_name, { 'inputs': inputs, 'parameters': { 'temperature': self.config.temperature }, 'options': { 'wait_for_model': True } })
        if isinstance(r, list) and len(r) and isinstance(r[0], dict) and 'generated_text' in r[0]:
            return r[0]['generated_text']
        if isinstance(r, str):
            return r
        if isinstance(r, dict) and 'generated_text' in r:
            return r['generated_text']
        return json.dumps(r)

    async def generate_stream(self, prompt: str, system_instruction: str = '') -> AsyncGenerator[dict, None]:
        text = await self.generate(prompt, system_instruction)
        yield { 'delta': text, 'finish_reason': None, 'usage': None }



import re
from typing import List, Dict, Any
from .config import RerankingConfig

class LLMReranker:
    def __init__(self, llm, config: RerankingConfig):
        self.llm = llm
        self.config = config

    async def rerank(self, query: str, documents: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        if not documents:
            return []
        
        scored = []
        for doc in documents:
            score = await self.score_document(query, doc['content'])
            doc_copy = doc.copy()
            doc_copy['score'] = score
            scored.append(doc_copy)
            
        scored.sort(key=lambda x: x['score'], reverse=True)
        return scored[:self.config.top_n]

    async def score_document(self, query: str, content: str) -> int:
        prompt = f'Analyze relevance (0-10) of document to query. Return ONLY integer.\nQuery: "{query}"\nDoc: "{content[:1000]}..."'
        try:
            res = await self.llm.generate(prompt)
            match = re.search(r'\d+', res)
            return int(match.group(0)) if match else 0
        except Exception:
            return 0



import argparse
import asyncio
import json
import os
from .config import VectraConfig
from .webconfig_server import start as start_webconfig

def load_config(path):
    with open(path, 'r', encoding='utf-8') as f:
        return VectraConfig.model_validate(json.load(f))

async def main_async():
    parser = argparse.ArgumentParser(prog='vectra')
    parser.add_argument('cmd', choices=['ingest','query','webconfig', 'dashboard'])
    parser.add_argument('target', nargs='?')
    parser.add_argument('--config')
    parser.add_argument('--stream', action='store_true')
    args = parser.parse_args()
    if args.cmd == 'webconfig':
        cfg_path = args.config or os.path.join(os.getcwd(), 'vectra-config.json')
        start_webconfig(cfg_path, 'webconfig')
        await asyncio.Event().wait()
        return

    if args.cmd == 'dashboard':
        cfg_path = args.config or os.path.join(os.getcwd(), 'vectra-config.json')
        start_webconfig(cfg_path, 'dashboard')
        await asyncio.Event().wait()
        return
    
    if not args.config:
        raise SystemExit('--config is required for ingest/query')
        
    # Lazy import to avoid hard dependencies for webconfig
    from .core import VectraClient
    
    cfg = load_config(args.config)
    client = VectraClient(cfg)
    if args.cmd == 'ingest':
        await client.ingest_documents(os.path.abspath(args.target))
        print('Ingestion complete')
    else:
        res = await client.query_rag(args.target, None, args.stream)
        if args.stream:
            async for chunk in res:
                t = chunk.get('delta') if isinstance(chunk, dict) else str(chunk)
                print(t, end='')
            print()
        else:
            print(json.dumps(res, ensure_ascii=False))

def main():
    asyncio.run(main_async())



import os
import json
import asyncio
import re
import math
from typing import List, Optional
from concurrent.futures import ThreadPoolExecutor

# Library imports (assumed available)
try:
    from pypdf import PdfReader
except ImportError:
    PdfReader = None
try:
    import mammoth
except ImportError:
    mammoth = None
try:
    import openpyxl
except ImportError:
    openpyxl = None

from .config import ChunkingStrategy, ChunkingConfig

class DocumentProcessor:
    def __init__(self, config: ChunkingConfig, agentic_llm=None):
        self.config = config
        self.agentic_llm = agentic_llm
        self._executor = ThreadPoolExecutor()
        self._last_pages = None

    async def load_document(self, file_path: str) -> str:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(self._executor, self._load_sync, file_path)

    def _load_sync(self, file_path: str) -> str:
        ext = os.path.splitext(file_path)[1].lower()
        if ext == '.pdf':
            if not PdfReader: raise ImportError("pypdf not installed")
            reader = PdfReader(file_path)
            pages = [page.extract_text() or "" for page in reader.pages]
            self._last_pages = pages
            return "\n".join(pages)
        
        elif ext == '.docx':
            if not mammoth: raise ImportError("mammoth not installed")
            with open(file_path, "rb") as docx_file:
                result = mammoth.extract_raw_text(docx_file)
                return result.value
        
        elif ext in ['.txt', '.md']:
            with open(file_path, 'r', encoding='utf-8') as f:
                return f.read()
        
        elif ext in ['.xlsx', '.xls']:
            if not openpyxl: raise ImportError("openpyxl not installed")
            wb = openpyxl.load_workbook(file_path, data_only=True)
            sheet = wb.active
            text = []
            for row in sheet.iter_rows(values_only=True):
                text.append(" ".join([str(c) for c in row if c is not None]))
            return "\n".join(text)
            
        raise ValueError(f"Unsupported file extension: {ext}")

    async def process(self, text: str) -> List[str]:
        if self.config.strategy == ChunkingStrategy.AGENTIC:
            return await self.agentic_split(text)
        return self.recursive_split(text)

    def _entropy(self, s: str) -> float:
        if not s:
            return 0.0
        freq = {}
        for ch in s:
            freq[ch] = freq.get(ch, 0) + 1
        length = len(s)
        H = 0.0
        for count in freq.values():
            p = count / length
            H += -p * math.log2(p)
        return H

    def recursive_split(self, text: str) -> List[str]:
        chunks = []
        # Match JS logic: max(500, config.chunkSize || 1000)
        size_chars = max(500, int(self.config.chunk_size) if self.config.chunk_size else 1000)
        # Match JS logic: max(0, config.chunkOverlap || 200)
        base_overlap = max(0, int(self.config.chunk_overlap) if self.config.chunk_overlap else 200)
        
        # Split by sentences (lookbehind for . ! ?)
        # JS: text.split(/(?<=[.!?])\s+/)
        sentences = re.split(r'(?<=[.!?])\s+', text)
        
        current = ''
        for s in sentences:
            candidate = f"{current} {s}" if current else s
            if len(candidate) >= size_chars:
                entropy = self._entropy(candidate)
                # overlap = Math.min(baseOverlap + Math.floor(entropy * 50), Math.floor(sizeChars / 3));
                overlap = min(base_overlap + math.floor(entropy * 50), math.floor(size_chars / 3))
                chunks.append(candidate)
                # current = candidate.slice(Math.max(0, candidate.length - overlap));
                start_slice = max(0, len(candidate) - overlap)
                current = candidate[start_slice:]
            else:
                current = candidate
        
        if current:
            chunks.append(current)
            
        return chunks

    def compute_chunk_metadata(self, file_path: str, raw_text: str, chunks: List[str]) -> List[dict]:
        ext = os.path.splitext(file_path)[1].lower()
        title = os.path.basename(file_path)
        positions = []
        cursor = 0
        for c in chunks:
            idx = raw_text.find(c, cursor)
            start = idx if idx >= 0 else 0
            end = start + len(c)
            positions.append((start, end))
            cursor = end
        pages_meta = None
        if ext == '.pdf' and isinstance(self._last_pages, list):
            lens = [len(p) for p in self._last_pages]
            cum = []
            acc = 0
            for l in lens:
                acc += l
                cum.append(acc)
            tmp = []
            for (start, end) in positions:
                pf = next((i for i, x in enumerate(cum) if x >= start), -1) + 1
                pt = next((i for i, x in enumerate(cum) if x >= end), -1) + 1
                tmp.append({'pageFrom': pf or 1, 'pageTo': pt or (pf or 1)})
            pages_meta = tmp
        sections = None
        if ext in ['.md', '.txt']:
            lines = raw_text.split('\n')
            heads = []
            offset = 0
            for ln in lines:
                if ln.strip().startswith('#'):
                    heads.append({'pos': offset, 'text': ln.strip().lstrip('#').strip()})
                offset += len(ln) + 1
            tmp = []
            for (start, _) in positions:
                cand = [h for h in heads if h['pos'] <= start]
                h = cand[-1]['text'] if cand else None
                tmp.append(h)
            sections = tmp
        out = []
        for i, (start, end) in enumerate(positions):
            m = {
                'fileType': ext,
                'docTitle': title,
                'chunkIndex': i,
                'pageFrom': pages_meta[i]['pageFrom'] if pages_meta else None,
                'pageTo': pages_meta[i]['pageTo'] if pages_meta else None,
                'section': sections[i] if sections else None
            }
            out.append(m)
        return out

    async def agentic_split(self, text: str) -> List[str]:
        if not self.agentic_llm:
            raise ValueError("Agentic LLM not configured.")
        
        windows = self.recursive_split(text)
        final_chunks = []
        
        for window in windows:
            prompt = f'Split this text into semantically complete propositions. Return a VALID JSON list of strings. Do not include Markdown formatting.\nText: "{window}"'
            try:
                response = await self.agentic_llm.generate(prompt)
                # Cleanup JSON
                clean_json = response.replace("```json", "").replace("```", "").strip()
                # Find the first [ and last ]
                start = clean_json.find('[')
                end = clean_json.rfind(']')
                if start != -1 and end != -1:
                    clean_json = clean_json[start:end+1]
                
                parsed = json.loads(clean_json)
                if isinstance(parsed, list):
                    final_chunks.extend(parsed)
                else:
                    final_chunks.append(window)
            except Exception:
                final_chunks.append(window)
                
        return final_chunks



from typing import List, Dict, Any, Optional, AsyncGenerator
import re
import hashlib
import asyncio
import os
import uuid
from .config import VectraConfig, ProviderType, ChunkingStrategy, RetrievalStrategy
from .observability import SQLiteLogger
from .processor import DocumentProcessor
from .backends.openai import OpenAIBackend
from .backends.gemini import GeminiBackend
from .backends.anthropic import AnthropicBackend
from .backends.openrouter import OpenRouterBackend
from .backends.prisma_store import PrismaVectorStore
from .backends.postgres_store import PostgresVectorStore
from .backends.chroma_store import ChromaVectorStore
from .backends.qdrant_store import QdrantVectorStore
from .backends.milvus_store import MilvusVectorStore
from .backends.huggingface import HuggingFaceBackend
from .reranker import LLMReranker
from .memory import InMemoryHistory, RedisHistory, PostgresHistory
from .backends.ollama import OllamaBackend

class VectraClient:
    def __init__(self, config: VectraConfig):
        self.config = config
        self.callbacks = config.callbacks or []
        self._embedding_cache: Dict[str, List[float]] = {}
        
        # Initialize observability
        self.logger = SQLiteLogger(config.observability)

        self.embedder = self._create_embedder()
        self.llm = self._create_llm(config.llm)
        
        if config.database.type == 'prisma':
            self.vector_store = PrismaVectorStore(config.database)
        elif config.database.type == 'postgres':
            self.vector_store = PostgresVectorStore(config.database)
        elif config.database.type == 'chroma':
            self.vector_store = ChromaVectorStore(config.database)
        elif config.database.type == 'qdrant':
            self.vector_store = QdrantVectorStore(config.database)
        elif config.database.type == 'milvus':
            self.vector_store = MilvusVectorStore(config.database)
        else:
            raise ValueError(f"Unsupported database type: {config.database.type}")
            
        agentic_llm = None
        if config.chunking.strategy == ChunkingStrategy.AGENTIC and config.chunking.agentic_llm:
            agentic_llm = self._create_llm(config.chunking.agentic_llm)
            
        self.processor = DocumentProcessor(config.chunking, agentic_llm)

        mm = int((getattr(self.config, 'memory', {}) or {}).get('max_messages', 20))
        mem = (getattr(self.config, 'memory', {}) or {})
        if mem.get('enabled'):
            if mem.get('type') == 'in-memory':
                self.history = InMemoryHistory(mm)
            elif mem.get('type') == 'redis':
                rc = mem.get('redis', {}) or {}
                self.history = RedisHistory(rc.get('client_instance'), rc.get('key_prefix', 'vectra:chat:'), mm)
            elif mem.get('type') == 'postgres':
                pc = mem.get('postgres', {}) or {}
                self.history = PostgresHistory(pc.get('client_instance'), pc.get('table_name', 'ChatMessage'), pc.get('column_map', { 'sessionId': 'sessionId', 'role': 'role', 'content': 'content', 'createdAt': 'createdAt' }), mm)
            else:
                self.history = None
        else:
            self.history = None
        
        if config.retrieval and config.retrieval.llm_config:
            self.retrieval_llm = self._create_llm(config.retrieval.llm_config)
            
        self.reranker = None
        if config.reranking and config.reranking.enabled:
            rerank_llm = self._create_llm(config.reranking.llm_config) if config.reranking.llm_config else self.llm
            self.reranker = LLMReranker(rerank_llm, config.reranking)

    def _create_embedder(self):
        conf = self.config.embedding
        if conf.provider == ProviderType.OPENAI: return OpenAIBackend(conf)
        if conf.provider == ProviderType.GEMINI: return GeminiBackend(conf)
        if conf.provider == ProviderType.OLLAMA: return OllamaBackend(conf)
        raise ValueError(f"Embedding provider {conf.provider} not implemented")

    def _create_llm(self, conf):
        if conf.provider == ProviderType.OPENAI: return OpenAIBackend(conf)
        if conf.provider == ProviderType.GEMINI: return GeminiBackend(conf)
        if conf.provider == ProviderType.ANTHROPIC: return AnthropicBackend(conf)
        if conf.provider == ProviderType.OPENROUTER: return OpenRouterBackend(conf)
        if conf.provider == ProviderType.HUGGINGFACE: return HuggingFaceBackend(conf)
        if conf.provider == ProviderType.OLLAMA: return OllamaBackend(conf)
        raise ValueError(f"LLM provider {conf.provider} not implemented")

    def _trigger(self, event: str, *args):
        for cb in self.callbacks:
            if hasattr(cb, event) and callable(getattr(cb, event)):
                getattr(cb, event)(*args)

    def _is_temporary_file(self, path: str) -> bool:
        name = os.path.basename(path)
        if name.startswith('~$'): return True
        if name.endswith('.tmp') or name.endswith('.temp'): return True
        if name.endswith('.crdownload') or name.endswith('.part'): return True
        if name.startswith('.'): return True
        return False
    
    async def ingest_documents(self, file_path: str, ingestion_mode: str = "append"):
        if os.path.isdir(file_path):
            summary = { 'processed': 0, 'succeeded': 0, 'failed': 0, 'errors': [] }
            for root, _, files in os.walk(file_path):
                for file in files:
                    full = os.path.join(root, file)
                    if self._is_temporary_file(full): 
                        continue
                    summary['processed'] += 1
                    try:
                        await self.ingest_documents(full, ingestion_mode=ingestion_mode)
                        summary['succeeded'] += 1
                    except Exception as e:
                        summary['failed'] += 1
                        summary['errors'].append({ 'file': full, 'message': str(e) })
                        self._trigger('on_error', e)
            self._trigger('on_ingest_summary', summary)
            return

        try:
            import time
            t0 = time.time()
            trace_id = str(uuid.uuid4())
            root_span_id = str(uuid.uuid4())
            provider = self.config.embedding.provider
            model_name = self.config.embedding.model_name
            
            self._trigger('on_ingest_start', file_path)
            
            abs_path = os.path.abspath(file_path)
            try:
                size = int(os.path.getsize(file_path))
                mtime = int(os.path.getmtime(file_path))
            except Exception:
                size = 0
                mtime = 0
            md5 = hashlib.md5()
            sha = hashlib.sha256()
            with open(file_path, 'rb') as f:
                while True:
                    b = f.read(8192)
                    if not b: break
                    md5.update(b)
                    sha.update(b)
            file_md5 = md5.hexdigest()
            file_sha256 = sha.hexdigest()
            validation = { 'absolutePath': abs_path, 'fileMD5': file_md5, 'fileSHA256': file_sha256, 'fileSize': size, 'lastModified': mtime, 'timestamp': int(time.time()*1000) }
            self._trigger('on_pre_ingestion_validation', validation)
            exists = False
            if hasattr(self.vector_store, 'file_exists'):
                try:
                    exists = await self.vector_store.file_exists(file_sha256, size, mtime)
                except Exception:
                    exists = False
            mode = str(ingestion_mode or "append").lower()
            if mode not in ("append", "skip", "replace", "upsert"):
                mode = "append"
            if exists and mode in ("append", "skip"):
                self._trigger('on_ingest_skipped', validation)
                return
            raw_text = await self.processor.load_document(file_path)
            
            self._trigger('on_chunking_start', self.config.chunking.strategy)
            chunks = await self.processor.process(raw_text)
            
            self._trigger('on_embedding_start', len(chunks))
            hashes = [hashlib.sha256(c.encode('utf-8')).hexdigest() for c in chunks]
            uncached = [chunks[i] for i,h in enumerate(hashes) if h not in self._embedding_cache]
            if len(uncached) > 0:
                ing = (getattr(self.config, 'ingestion', {}) or {})
                enabled = bool(ing.get('rate_limit_enabled', False))
                default_limit = int(ing.get('concurrency_limit', 5))
                limit = default_limit if enabled else len(uncached)
                new_embeds: List[List[float]] = []
                for i in range(0, len(uncached), limit):
                    batch = uncached[i:i+limit]
                    attempt = 0
                    delay = 0.5
                    while True:
                        try:
                            out = await self.embedder.embed_documents(batch)
                            new_embeds.extend(out)
                            break
                        except Exception as e:
                            attempt += 1
                            if attempt >= 3:
                                raise e
                            await asyncio.sleep(delay)
                            delay = min(4.0, delay * 2)
                j = 0
                for i,h in enumerate(hashes):
                    if h not in self._embedding_cache:
                        self._embedding_cache[h] = new_embeds[j]
                        j += 1
            embeddings = [self._embedding_cache[h] for h in hashes]
            metas = self.processor.compute_chunk_metadata(file_path, raw_text, chunks)
            documents = []
            for i, content in enumerate(chunks):
                md = metas[i] if i < len(metas) else {}
                meta = {
                    'source': file_path,
                    'absolutePath': abs_path,
                    'fileMD5': file_md5,
                    'fileSHA256': file_sha256,
                    'fileSize': size,
                    'lastModified': mtime,
                    'chunk_index': i,
                    'sha256': hashes[i],
                    'fileType': md.get('fileType'),
                    'docTitle': md.get('docTitle'),
                    'pageFrom': md.get('pageFrom'),
                    'pageTo': md.get('pageTo'),
                    'section': md.get('section')
                }
                # Filter None values for ChromaDB compatibility
                meta = {k: v for k, v in meta.items() if v is not None}
                
                documents.append({
                    'content': content,
                    'embedding': embeddings[i],
                    'metadata': meta
                })

            if getattr(self.config, 'metadata', None) and self.config.metadata.get('enrichment'): 
                import json
                enriched = []
                for c in chunks:
                    try:
                        prompt = f"Summarize and extract keywords and questions from the following text. Return STRICT JSON with keys: summary (string), keywords (array of strings), hypothetical_questions (array of strings).\nText:\n{c}"
                        out = await self.llm.generate(prompt, 'You return valid JSON only.')
                        clean = str(out).replace('```json','').replace('```','').strip()
                        p = json.loads(clean)
                        enriched.append({ 'summary': p.get('summary',''), 'keywords': p.get('keywords', []), 'hypothetical_questions': p.get('hypothetical_questions', []) })
                    except Exception:
                        import re
                        words = re.findall(r"[a-zA-Z0-9]+", c.lower())
                        freq = {}
                        for w in words:
                            if len(w) > 3:
                                freq[w] = freq.get(w, 0) + 1
                        top = [w for w,_ in sorted(freq.items(), key=lambda kv: kv[1], reverse=True)[:10]]
                        enriched.append({ 'summary': c[:300], 'keywords': top, 'hypothetical_questions': [] })
                for i in range(len(documents)):
                    documents[i]['metadata']['summary'] = enriched[i]['summary']
                    documents[i]['metadata']['keywords'] = enriched[i]['keywords']
                    documents[i]['metadata']['hypothetical_questions'] = enriched[i]['hypothetical_questions']
                
            if hasattr(self.vector_store, 'ensure_indexes'):
                try:
                    await self.vector_store.ensure_indexes()
                except Exception:
                    pass
            exists_server = False
            if hasattr(self.vector_store, 'file_exists'):
                try:
                    exists_server = await self.vector_store.file_exists(file_sha256, size, mtime)
                except Exception:
                    exists_server = False
            if exists_server and mode in ("append", "skip"):
                self._trigger('on_ingest_skipped', validation)
                return

            if mode == "replace":
                try:
                    await self.vector_store.delete_documents({ "absolutePath": abs_path })
                except Exception:
                    pass

            if mode == "upsert":
                to_add = []
                for d in documents:
                    try:
                        updated = 0
                        if hasattr(self.vector_store, "update_documents"):
                            updated = await self.vector_store.update_documents(
                                { "sha256": d["metadata"].get("sha256") },
                                { "content": d["content"], "metadata": d["metadata"] },
                            )
                        if not updated and hasattr(self.vector_store, "delete_documents"):
                            await self.vector_store.delete_documents({ "sha256": d["metadata"].get("sha256") })
                        if not updated:
                            to_add.append(d)
                    except Exception:
                        to_add.append(d)
                documents = to_add
                if not documents:
                    duration_ms = int((time.time() - t0) * 1000)
                    self._trigger('on_ingest_end', file_path, 0, duration_ms)
                    return
            attempt = 0
            delay = 0.5
            while True:
                try:
                    await self.vector_store.add_documents(documents)
                    break
                except Exception as e:
                    attempt += 1
                    if attempt >= 3:
                        raise e
                    await asyncio.sleep(delay)
                    delay = min(4.0, delay * 2)
            duration_ms = int((time.time() - t0) * 1000)
            self._trigger('on_ingest_end', file_path, len(chunks), duration_ms)
            
            self.logger.log_trace({
                'trace_id': trace_id,
                'span_id': root_span_id,
                'name': 'ingest_documents',
                'start_time': int(t0 * 1000),
                'end_time': int(time.time() * 1000),
                'input': {'file_path': file_path},
                'output': {'chunks': len(chunks), 'duration_ms': duration_ms},
                'attributes': {'file_size': size},
                'provider': provider,
                'model_name': model_name
            })
            self.logger.log_metric({'name': 'ingest_latency', 'value': duration_ms, 'tags': {'type': 'single_file'}})
            
        except Exception as e:
            self._trigger('on_error', e)
            if 'trace_id' in locals():
                 self.logger.log_trace({
                    'trace_id': trace_id,
                    'span_id': root_span_id,
                    'name': 'ingest_documents',
                    'start_time': int(t0 * 1000),
                    'end_time': int(time.time() * 1000),
                    'input': {'file_path': file_path},
                    'error': {'message': str(e)},
                    'status': 'error',
                    'provider': provider,
                    'model_name': model_name
                })
            raise e

    async def list_documents(self, filter: Optional[Dict[str, Any]] = None, limit: int = 100) -> List[Dict[str, Any]]:
        return await self.vector_store.list_documents(filter=filter, limit=limit)

    async def delete_documents(self, filter: Dict[str, Any]) -> int:
        return await self.vector_store.delete_documents(filter)

    async def update_documents(self, filter: Dict[str, Any], update_data: Dict[str, Any]) -> int:
        return await self.vector_store.update_documents(filter, update_data)

    async def _generate_hyde_query(self, query: str) -> str:
        prompt = f'Please write a plausible passage that answers the question: "{query}".'
        return await self.retrieval_llm.generate(prompt)

    async def _generate_multi_queries(self, query: str) -> List[str]:
        prompt = f'Generate 3 different versions of the user question to retrieve relevant documents. Return them separated by newlines.\nOriginal: {query}'
        response = await self.retrieval_llm.generate(prompt)
        return [line.strip() for line in response.split('\n') if line.strip()][:3]

    async def _generate_hypothetical_questions(self, query: str) -> List[str]:
        out = await self.retrieval_llm.generate(f"Generate 3 hypothetical questions related to the query. Return a VALID JSON array of strings.\nQuery: {query}")
        try:
            import json
            arr = json.loads(str(out).strip().replace('```json','').replace('```',''))
            return arr[:3] if isinstance(arr, list) else []
        except Exception:
            return []

    def _token_estimate(self, text: str) -> int:
        return max(1, (len(text or '') + 3) // 4)

    def _build_context_parts(self, docs: List[Dict[str, Any]], query: str) -> List[str]:
        budget = int(self.config.query_planning.get('token_budget', 2048)) if getattr(self.config, 'query_planning', None) else 2048
        prefer_summ = int(self.config.query_planning.get('prefer_summaries_below', 1024)) if getattr(self.config, 'query_planning', None) else 1024
        parts: List[str] = []
        used = 0
        for d in docs:
            md = d.get('metadata', {})
            t = md.get('docTitle') or ''
            sec = md.get('section') or ''
            pf = md.get('pageFrom'); pt = md.get('pageTo')
            pages = f"pages {pf}-{pt}" if pf and pt else ''
            summ = md.get('summary') or d.get('content','')[:800]
            chosen = summ if self._token_estimate(summ) <= prefer_summ else d.get('content','')[:1200]
            part = f"{t} {sec} {pages}\n{chosen}"
            est = self._token_estimate(part)
            if used + est > budget:
                break
            parts.append(part)
            used += est
        return parts

    def _extract_snippets(self, docs: List[Dict[str, Any]], query: str, max_snippets: int) -> List[str]:
        terms = [t for t in re.findall(r"[a-zA-Z0-9]+", query.lower()) if len(t) > 2]
        out: List[str] = []
        for d in docs:
            sents = re.split(r"(?<=[.!?])\s+", d.get('content',''))
            for s in sents:
                l = s.lower()
                score = sum(1 for t in terms if t in l)
                if score > 0:
                    md = d.get('metadata', {})
                    pf = md.get('pageFrom'); pt = md.get('pageTo')
                    pages = f"pages {pf}-{pt}" if pf and pt else ''
                    out.append(f"{md.get('docTitle') or ''} {md.get('section') or ''} {pages}\n{s}")
                    if len(out) >= max_snippets:
                        return out
        return out

    def _reciprocal_rank_fusion(self, doc_lists: List[List[Dict]], k=60) -> List[Dict]:
        """Simple RRF implementation for merging results."""
        scores = {}
        content_map = {}
        
        for doc_list in doc_lists:
            for rank, doc in enumerate(doc_list):
                content = doc['content']
                if content not in content_map:
                    content_map[content] = doc
                if content not in scores:
                    scores[content] = 0
                scores[content] += 1 / (k + rank + 1)
        
        sorted_content = sorted(scores.keys(), key=lambda x: scores[x], reverse=True)
        return [content_map[c] for c in sorted_content]

    def _mmr_select(self, candidates: List[Dict[str, Any]], k: int, mmr_lambda: float) -> List[Dict[str, Any]]:
        if not candidates:
            return []
        k_int = max(1, int(k))
        lam = max(0.0, min(1.0, float(mmr_lambda)))

        def tokens(text: str) -> set:
            return set(t for t in re.findall(r"[a-zA-Z0-9]+", (text or "").lower()) if len(t) > 2)

        cand = []
        for d in candidates:
            dd = dict(d)
            dd["_tokens"] = tokens(dd.get("content", ""))
            dd["_rel"] = float(dd.get("score", 0.0) or 0.0)
            cand.append(dd)

        cand.sort(key=lambda x: x.get("_rel", 0.0), reverse=True)
        selected: List[Dict[str, Any]] = []
        selected_tokens: List[set] = []

        first = cand.pop(0)
        selected.append(first)
        selected_tokens.append(first.get("_tokens") or set())

        def jaccard(a: set, b: set) -> float:
            if not a or not b:
                return 0.0
            inter = len(a & b)
            if inter == 0:
                return 0.0
            union = len(a | b)
            return inter / union if union else 0.0

        while cand and len(selected) < k_int:
            best_idx = -1
            best_score = None
            for i, d in enumerate(cand):
                rel = d.get("_rel", 0.0)
                dt = d.get("_tokens") or set()
                div = 0.0
                for st in selected_tokens:
                    div = max(div, jaccard(dt, st))
                score = lam * rel - (1.0 - lam) * div
                if best_score is None or score > best_score:
                    best_score = score
                    best_idx = i
            if best_idx < 0:
                break
            picked = cand.pop(best_idx)
            selected.append(picked)
            selected_tokens.append(picked.get("_tokens") or set())

        out = []
        for d in selected[:k_int]:
            dd = dict(d)
            dd.pop("_tokens", None)
            dd.pop("_rel", None)
            out.append(dd)
        return out

    async def query_rag(self, query: str, filter: Optional[Dict] = None, stream: bool = False, session_id: Optional[str] = None) -> Dict[str, Any] | AsyncGenerator[str, None]:
        trace_id = str(uuid.uuid4())
        root_span_id = str(uuid.uuid4())
        
        provider = self.config.llm.provider
        model_name = self.config.llm.model_name

        if session_id:
             self.logger.update_session(session_id, None, {'last_query': query})

        try:
            import time
            t_start = time.time()
            t_ret = time.time()
            self._trigger('on_retrieval_start', query)
            
            strategy = self.config.retrieval.strategy
            docs = []
            k = self.config.reranking.window_size if (self.config.reranking and self.config.reranking.enabled) else 5
            
            query_vector = await self.embedder.embed_query(query)

            if strategy == RetrievalStrategy.HYDE:
                hypothetical_doc = await self._generate_hyde_query(query)
                hyde_vector = await self.embedder.embed_query(hypothetical_doc)
                docs = await self.vector_store.similarity_search(hyde_vector, k, filter)
                
            elif strategy == RetrievalStrategy.MULTI_QUERY:
                queries = await self._generate_multi_queries(query)
                if getattr(self.config, 'query_planning', None):
                    hyps = await self._generate_hypothetical_questions(query)
                    queries.extend(hyps)
                queries.append(query)
                result_lists = []
                for q in queries:
                    vec = await self.embedder.embed_query(q)
                    result_lists.append(await self.vector_store.similarity_search(vec, k, filter))
                docs = self._reciprocal_rank_fusion(result_lists, k=60)
                
            elif strategy == RetrievalStrategy.HYBRID:
                 # Hybrid = Vector Search + (Simulated) Keyword Search via specialized Store method
                 docs = await self.vector_store.hybrid_search(query, query_vector, k, filter)

            elif strategy == RetrievalStrategy.MMR:
                fetch_k = int(getattr(self.config.retrieval, "mmr_fetch_k", 20))
                mmr_lam = float(getattr(self.config.retrieval, "mmr_lambda", 0.5))
                candidates = await self.vector_store.similarity_search(query_vector, max(fetch_k, k), filter)
                docs = self._mmr_select(candidates, k, mmr_lam)

            else: # NAIVE
                docs = await self.vector_store.similarity_search(query_vector, k, filter)
                
            if self.config.reranking and self.config.reranking.enabled and self.reranker:
                self._trigger('on_reranking_start', len(docs))
                docs = await self.reranker.rerank(query, docs)
                self._trigger('on_reranking_end', len(docs))
                
            self._trigger('on_retrieval_end', len(docs), int((time.time() - t_ret) * 1000))
            
            embedding_provider = self.config.embedding.provider
            embedding_model_name = self.config.embedding.model_name
            self.logger.log_trace({
                'trace_id': trace_id,
                'span_id': str(uuid.uuid4()),
                'parent_span_id': root_span_id,
                'name': 'retrieval',
                'start_time': int(t_ret * 1000),
                'end_time': int(time.time() * 1000),
                'input': {'query': query, 'filter': filter, 'strategy': strategy},
                'output': {'documents_found': len(docs)},
                'provider': embedding_provider,
                'model_name': embedding_model_name
            })

            terms = [t for t in re.findall(r"[a-zA-Z0-9]+", query.lower()) if len(t) > 2]
            boosted = []
            for d in docs:
                kws = [str(k).lower() for k in (d.get('metadata',{}).get('keywords') or [])]
                match = sum(1 for t in terms if t in kws)
                dd = dict(d)
                dd['_boost'] = match
                boosted.append(dd)
            boosted.sort(key=lambda x: x.get('_boost',0), reverse=True)
            context_parts = self._build_context_parts(boosted, query)
            if getattr(self.config, 'grounding', None) and self.config.grounding.get('enabled'):
                max_snippets = int(self.config.grounding.get('max_snippets', 3))
                snippets = self._extract_snippets(boosted, query, max_snippets)
                if self.config.grounding.get('strict'):
                    context_parts = snippets
                else:
                    context_parts.extend(snippets)
            context = "\n---\n".join(context_parts)
            import inspect
            history_text = ""
            if self.history and session_id:
                get_recent = getattr(self.history, 'get_recent', None)
                if callable(get_recent):
                    if inspect.iscoroutinefunction(get_recent):
                        recent = await get_recent(session_id, int((getattr(self.config, 'memory', {}) or {}).get('max_messages', 10)))
                    else:
                        recent = get_recent(session_id, int((getattr(self.config, 'memory', {}) or {}).get('max_messages', 10)))
                    history_text = "\n".join([f"{m['role'].upper()}: {m['content']}" for m in recent])
            if getattr(self.config, 'prompts', None) and self.config.prompts.get('query'):
                prompt = str(self.config.prompts.get('query')).replace('{{context}}', context).replace('{{question}}', query)
                if history_text:
                    prompt = f"Conversation:\n{history_text}\n\n{prompt}"
            else:
                conv_section = f"Conversation:\n{history_text}\n\n" if history_text else ""
                prompt = f"Answer the question using the provided summaries and cite titles/sections/pages where relevant.\nContext:\n{context}\n\n{conv_section}Question: {query}"
            
            t_gen = time.time()
            self._trigger('on_generation_start', prompt)
            
            system_inst = "You are a helpful RAG assistant."
            if stream:
                # Streaming Logic
                self.logger.log_trace({
                    'trace_id': trace_id,
                    'span_id': str(uuid.uuid4()),
                    'parent_span_id': root_span_id,
                    'name': 'generation_stream_start',
                    'start_time': int(t_gen * 1000),
                    'end_time': int(time.time() * 1000),
                    'input': {'prompt': prompt},
                    'output': {'stream': True},
                    'provider': provider,
                    'model_name': model_name
                })
                return self.llm.generate_stream(prompt, system_inst) # Need to implement this in LLMs
            else:
                answer = await self.llm.generate(prompt, system_inst)
                if self.history and session_id:
                    add_msg = getattr(self.history, 'add_message', None)
                    if callable(add_msg):
                        if inspect.iscoroutinefunction(add_msg):
                            await add_msg(session_id, 'user', query)
                            await add_msg(session_id, 'assistant', str(answer))
                        else:
                            add_msg(session_id, 'user', query)
                            add_msg(session_id, 'assistant', str(answer))
                
                gen_ms = int((time.time() - t_gen) * 1000)
                self._trigger('on_generation_end', answer, gen_ms)
                
                prompt_chars = len(prompt)
                answer_chars = len(str(answer))

                self.logger.log_trace({
                    'trace_id': trace_id,
                    'span_id': str(uuid.uuid4()),
                    'parent_span_id': root_span_id,
                    'name': 'generation',
                    'start_time': int(t_gen * 1000),
                    'end_time': int(time.time() * 1000),
                    'input': {'prompt': prompt},
                    'output': {'answer': str(answer)[:1000]},
                    'attributes': {'prompt_chars': prompt_chars, 'completion_chars': answer_chars},
                    'provider': provider,
                    'model_name': model_name
                })
                
                self.logger.log_metric({'name': 'prompt_chars', 'value': prompt_chars})
                self.logger.log_metric({'name': 'completion_chars', 'value': answer_chars})

                self.logger.log_trace({
                    'trace_id': trace_id,
                    'span_id': root_span_id,
                    'name': 'query_rag',
                    'start_time': int(t_start * 1000),
                    'end_time': int(time.time() * 1000),
                    'input': {'query': query, 'session_id': session_id},
                    'output': {'success': True},
                    'attributes': {'retrieval_ms': int((t_gen - t_ret) * 1000), 'gen_ms': gen_ms, 'doc_count': len(docs)},
                    'provider': provider,
                    'model_name': model_name
                })
                self.logger.log_metric({'name': 'query_latency', 'value': int((time.time() - t_start) * 1000), 'tags': {'type': 'total'}})

                if getattr(self.config, 'generation', None) and self.config.generation.get('output_format') == 'json':
                    try:
                        import json
                        parsed = json.loads(str(answer))
                        return {'answer': parsed, 'sources': [d['metadata'] for d in docs]}
                    except Exception:
                        return {'answer': answer, 'sources': [d['metadata'] for d in docs]}
                return {'answer': answer, 'sources': [d['metadata'] for d in docs]}
            
        except Exception as e:
            self._trigger('on_error', e)
            if 'trace_id' in locals():
                self.logger.log_trace({
                    'trace_id': trace_id,
                    'span_id': root_span_id,
                    'name': 'query_rag',
                    'start_time': int(t_start * 1000) if 't_start' in locals() else int(time.time() * 1000),
                    'end_time': int(time.time() * 1000),
                    'input': {'query': query},
                    'error': {'message': str(e)},
                    'status': 'error'
                })
            raise e

    async def evaluate(self, test_set: List[Dict[str, str]]) -> List[Dict[str, Any]]:
        report: List[Dict[str, Any]] = []
        for item in test_set:
            res = await self.query_rag(item['question'])
            ctx = "\n".join([s.get('summary','') for s in res.get('sources', [])])
            fp = f"Rate 0-1: Is the following Answer derived only from the Context?\nContext:\n{ctx}\n\nAnswer:\n{res.get('answer') if isinstance(res.get('answer'), str) else str(res.get('answer'))}"
            rp = f"Rate 0-1: Does the Answer correctly answer the Question?\nQuestion:\n{item['question']}\n\nAnswer:\n{res.get('answer') if isinstance(res.get('answer'), str) else str(res.get('answer'))}"
            faith = 0.0; rel = 0.0
            try:
                faith = max(0.0, min(1.0, float(str(await self.llm.generate(fp, 'You return a single number between 0 and 1.')))))
            except Exception:
                faith = 0.0
            try:
                rel = max(0.0, min(1.0, float(str(await self.llm.generate(rp, 'You return a single number between 0 and 1.')))))
            except Exception:
                rel = 0.0
            report.append({ 'question': item['question'], 'expectedGroundTruth': item.get('expectedGroundTruth',''), 'faithfulness': faith, 'relevance': rel })
        return report



import os
from typing import List, AsyncGenerator
from openai import AsyncOpenAI
from ..config import LLMConfig

class OpenRouterBackend:
    def __init__(self, config: LLMConfig):
        self.config = config
        api_key = config.api_key or os.getenv('OPENROUTER_API_KEY')
        if not api_key:
            raise ValueError('OpenRouter API Key missing. Set OPENROUTER_API_KEY.')
        self.client = AsyncOpenAI(
            api_key=api_key,
            base_url=config.base_url or 'https://openrouter.ai/api/v1',
            default_headers=config.default_headers or {
                'HTTP-Referer': os.getenv('OPENROUTER_REFERER', ''),
                'X-Title': os.getenv('OPENROUTER_TITLE', '')
            }
        )

    async def embed_documents(self, texts: List[str]) -> List[List[float]]:
        raise NotImplementedError('OpenRouter embeddings are not supported via this adapter.')

    async def embed_query(self, text: str) -> List[float]:
        raise NotImplementedError('OpenRouter embeddings are not supported via this adapter.')

    async def generate(self, prompt: str, system_instruction: str = 'You are a helpful assistant.') -> str:
        res = await self.client.chat.completions.create(
            model=self.config.model_name,
            messages=[
                { 'role': 'system', 'content': system_instruction },
                { 'role': 'user', 'content': prompt }
            ],
            temperature=self.config.temperature,
            max_tokens=self.config.max_tokens
        )
        return res.choices[0].message.content

    async def generate_stream(self, prompt: str, system_instruction: str = 'You are a helpful assistant.') -> AsyncGenerator[dict, None]:
        stream = await self.client.chat.completions.create(
            model=self.config.model_name,
            messages=[
                { 'role': 'system', 'content': system_instruction },
                { 'role': 'user', 'content': prompt }
            ],
            temperature=self.config.temperature,
            max_tokens=self.config.max_tokens,
            stream=True
        )
        async for chunk in stream:
            if chunk.choices[0].delta.content:
                yield { 'delta': chunk.choices[0].delta.content, 'finish_reason': None, 'usage': None }



import sqlite3
import json
import uuid
import time
import os

class SQLiteLogger:
    def __init__(self, config):
        self.enabled = config.enabled
        if not self.enabled:
            return

        self.project_id = config.project_id
        self.track_metrics = config.track_metrics
        self.track_traces = config.track_traces
        self.track_logs = config.track_logs
        self.session_tracking = config.session_tracking
        
        try:
            self.db_path = config.sqlite_path
            self._init_schema()
        except Exception as e:
            print(f"Failed to initialize SQLite logger: {e}")
            self.enabled = False

    def _init_schema(self):
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.executescript("""
                CREATE TABLE IF NOT EXISTS traces (
                    id TEXT PRIMARY KEY,
                    project_id TEXT,
                    trace_id TEXT,
                    span_id TEXT,
                    parent_span_id TEXT,
                    name TEXT,
                    start_time INTEGER,
                    end_time INTEGER,
                    duration INTEGER,
                    status TEXT,
                    attributes TEXT,
                    input TEXT,
                    output TEXT,
                    error TEXT,
                    provider TEXT,
                    model_name TEXT
                );
                CREATE TABLE IF NOT EXISTS metrics (
                    id TEXT PRIMARY KEY,
                    project_id TEXT,
                    name TEXT,
                    value REAL,
                    timestamp INTEGER,
                    tags TEXT
                );
                CREATE TABLE IF NOT EXISTS logs (
                    id TEXT PRIMARY KEY,
                    project_id TEXT,
                    level TEXT,
                    message TEXT,
                    timestamp INTEGER,
                    context TEXT
                );
                CREATE TABLE IF NOT EXISTS sessions (
                    id TEXT PRIMARY KEY,
                    project_id TEXT,
                    session_id TEXT,
                    user_id TEXT,
                    start_time INTEGER,
                    last_activity_time INTEGER,
                    metadata TEXT
                );
            """)
            
            # Migration: Add new columns if they don't exist
            try:
                cursor.execute("ALTER TABLE traces ADD COLUMN provider TEXT")
            except sqlite3.OperationalError:
                pass
            
            try:
                cursor.execute("ALTER TABLE traces ADD COLUMN model_name TEXT")
            except sqlite3.OperationalError:
                pass

    def log_trace(self, trace):
        if not self.enabled or not self.track_traces:
            return
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute("""
                    INSERT INTO traces (id, project_id, trace_id, span_id, parent_span_id, name, start_time, end_time, duration, status, attributes, input, output, error, provider, model_name)
                    VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
                """, (
                    str(uuid.uuid4()),
                    self.project_id,
                    trace.get('trace_id'),
                    trace.get('span_id'),
                    trace.get('parent_span_id'),
                    trace.get('name'),
                    trace.get('start_time'),
                    trace.get('end_time'),
                    (trace.get('end_time', 0) - trace.get('start_time', 0)),
                    trace.get('status', 'ok'),
                    json.dumps(trace.get('attributes', {})),
                    json.dumps(trace.get('input', {})),
                    json.dumps(trace.get('output', {})),
                    json.dumps(trace.get('error')) if trace.get('error') else None,
                    trace.get('provider'),
                    trace.get('model_name')
                ))
        except Exception as e:
            print(f"Failed to log trace: {e}")

    def log_metric(self, metric):
        if not self.enabled or not self.track_metrics:
            return
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute("""
                    INSERT INTO metrics (id, project_id, name, value, timestamp, tags)
                    VALUES (?, ?, ?, ?, ?, ?)
                """, (
                    str(uuid.uuid4()),
                    self.project_id,
                    metric.get('name'),
                    metric.get('value'),
                    int(time.time() * 1000),
                    json.dumps(metric.get('tags', {}))
                ))
        except Exception as e:
            print(f"Failed to log metric: {e}")

    def update_session(self, session_id, user_id=None, metadata=None):
        if not self.enabled or not self.session_tracking:
            return
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute("SELECT id FROM sessions WHERE session_id = ?", (session_id,))
                existing = cursor.fetchone()
                
                now = int(time.time() * 1000)
                if existing:
                    cursor.execute("""
                        UPDATE sessions 
                        SET last_activity_time = ?, metadata = ?
                        WHERE session_id = ?
                    """, (now, json.dumps(metadata or {}), session_id))
                else:
                    cursor.execute("""
                        INSERT INTO sessions (id, project_id, session_id, user_id, start_time, last_activity_time, metadata)
                        VALUES (?, ?, ?, ?, ?, ?, ?)
                    """, (
                        str(uuid.uuid4()),
                        self.project_id,
                        session_id,
                        user_id,
                        now,
                        now,
                        json.dumps(metadata or {})
                    ))
        except Exception as e:
            print(f"Failed to update session: {e}")



from enum import Enum
from typing import List, Optional, Any, Dict
from pydantic import BaseModel, Field, field_validator

class ProviderType(str, Enum):
    OPENAI = 'openai'
    ANTHROPIC = 'anthropic'
    GEMINI = 'gemini'
    OPENROUTER = 'openrouter'
    HUGGINGFACE = 'huggingface'
    OLLAMA = 'ollama'

class ChunkingStrategy(str, Enum):
    RECURSIVE = 'recursive'
    AGENTIC = 'agentic'

class RetrievalStrategy(str, Enum):
    NAIVE = 'naive'
    HYDE = 'hyde'
    MULTI_QUERY = 'multi_query'
    HYBRID = 'hybrid'  # New Strategy
    MMR = 'mmr'

class EmbeddingConfig(BaseModel):
    provider: ProviderType
    api_key: Optional[str] = None
    model_name: str = 'text-embedding-3-small'
    dimensions: Optional[int] = None

class LLMConfig(BaseModel):
    provider: ProviderType
    api_key: Optional[str] = None
    model_name: str
    temperature: float = 0.0
    max_tokens: int = 1024
    base_url: Optional[str] = None
    default_headers: Optional[Dict[str, str]] = None

class ChunkingConfig(BaseModel):
    strategy: ChunkingStrategy = ChunkingStrategy.RECURSIVE
    chunk_size: int = 1000
    chunk_overlap: int = 200
    separators: List[str] = ['\n\n', '\n', ' ', '']
    agentic_llm: Optional[LLMConfig] = None

    @field_validator('agentic_llm')
    def check_agentic_llm(cls, v, info):
        if info.data.get('strategy') == ChunkingStrategy.AGENTIC and not v:
            raise ValueError('agentic_llm required for AGENTIC strategy')
        return v

class RerankingConfig(BaseModel):
    enabled: bool = False
    provider: str = 'llm'
    llm_config: Optional[LLMConfig] = None
    top_n: int = 5
    window_size: int = 20

class RetrievalConfig(BaseModel):
    strategy: RetrievalStrategy = RetrievalStrategy.NAIVE
    llm_config: Optional[LLMConfig] = None
    hybrid_alpha: float = 0.5 # Alpha 0-1 (0 = keyword, 1 = dense)
    mmr_lambda: float = 0.5
    mmr_fetch_k: int = 20

    @field_validator('llm_config')
    def check_llm_config(cls, v, info):
        strategy = info.data.get('strategy')
        if strategy in [RetrievalStrategy.HYDE, RetrievalStrategy.MULTI_QUERY] and not v:
            raise ValueError('llm_config required for advanced retrieval')
        return v

class DatabaseConfig(BaseModel):
    type: str # 'prisma', 'chroma', 'custom'
    table_name: Optional[str] = None
    column_map: Optional[Dict[str, str]] = {"content": "content", "vector": "vector", "metadata": "metadata"}
    client_instance: Any  # Prisma client, Chroma client, etc.

class ObservabilityConfig(BaseModel):
    enabled: bool = False
    sqlite_path: str = 'vectra-observability.db'
    project_id: str = 'default'
    track_metrics: bool = True
    track_traces: bool = True
    track_logs: bool = True
    session_tracking: bool = True

class VectraConfig(BaseModel):
    embedding: EmbeddingConfig
    llm: LLMConfig
    database: DatabaseConfig
    chunking: ChunkingConfig = Field(default_factory=ChunkingConfig)
    retrieval: RetrievalConfig = Field(default_factory=RetrievalConfig)
    reranking: RerankingConfig = Field(default_factory=RerankingConfig)
    callbacks: List[Any] = []
    metadata: Optional[Dict[str, Any]] = None
    ingestion: Optional[Dict[str, Any]] = Field(default_factory=lambda: { 'rate_limit_enabled': False, 'concurrency_limit': 5 })
    memory: Optional[Dict[str, Any]] = Field(default_factory=lambda: { 'enabled': False, 'type': 'in-memory', 'max_messages': 20 })
    query_planning: Optional[Dict[str, Any]] = None
    grounding: Optional[Dict[str, Any]] = None
    generation: Optional[Dict[str, Any]] = None
    prompts: Optional[Dict[str, Any]] = None
    tracing: Optional[Dict[str, Any]] = None
    observability: ObservabilityConfig = Field(default_factory=ObservabilityConfig)



from typing import List, Dict, Any, Optional
from ..interfaces import VectorStore

class QdrantVectorStore(VectorStore):
    def __init__(self, config):
        self.config = config
        self.client = config.client_instance
        self.collection = config.table_name or 'rag_collection'

    async def add_documents(self, documents: List[Dict[str, Any]]):
        points = [{
            'id': i,
            'vector': doc['embedding'],
            'payload': {'content': doc['content'], 'metadata': doc['metadata']}
        } for i, doc in enumerate(documents)]
        await self.client.upsert(self.collection, points=points)

    def _normalize_filter(self, filter: Optional[Dict[str, Any]]):
        if not filter:
            return None
        if not isinstance(filter, dict):
            return filter
        if any(k in filter for k in ("must", "should", "must_not")):
            return filter
        must = []
        for k, v in filter.items():
            if isinstance(v, (str, int, float, bool)):
                must.append({"key": f"metadata.{k}", "match": {"value": v}})
        return {"must": must} if must else None

    async def similarity_search(self, vector: List[float], limit: int = 5, filter: Optional[Dict] = None) -> List[Dict[str, Any]]:
        q_filter = self._normalize_filter(filter)
        res = await self.client.search(self.collection, vector=vector, limit=limit, filter=q_filter)
        return [{'content': r['payload']['content'], 'metadata': r['payload'].get('metadata'), 'score': r.get('score', 1.0)} for r in res]

    async def hybrid_search(self, text: str, vector: List[float], limit: int = 5, filter: Optional[Dict] = None) -> List[Dict[str, Any]]:
        return await self.similarity_search(vector, limit, filter)

    async def _scroll_points(self, filter: Optional[Dict[str, Any]], limit: int, offset=None):
        if hasattr(self.client, "scroll"):
            q_filter = self._normalize_filter(filter)
            try:
                return await self.client.scroll(
                    self.collection,
                    scroll_filter=q_filter,
                    limit=limit,
                    with_payload=True,
                    with_vectors=False,
                    offset=offset,
                )
            except TypeError:
                return await self.client.scroll(
                    self.collection,
                    filter=q_filter,
                    limit=limit,
                    with_payload=True,
                    with_vectors=False,
                    offset=offset,
                )
        raise NotImplementedError("Qdrant client does not support scroll()")

    async def list_documents(self, filter: Optional[Dict[str, Any]] = None, limit: int = 100) -> List[Dict[str, Any]]:
        limit_int = max(1, int(limit))
        batch = min(256, limit_int)
        out: List[Dict[str, Any]] = []
        offset = None
        while len(out) < limit_int:
            res = await self._scroll_points(filter, batch, offset=offset)
            if isinstance(res, tuple) and len(res) == 2:
                points, new_offset = res
            else:
                points = res
                new_offset = None
            for p in points or []:
                payload = p.get("payload") or {}
                out.append(
                    {
                        "id": p.get("id"),
                        "content": payload.get("content", ""),
                        "metadata": payload.get("metadata") or {},
                    }
                )
                if len(out) >= limit_int:
                    break
            if not points or new_offset is None or new_offset == offset:
                break
            offset = new_offset
        return out

    async def delete_documents(self, filter: Dict[str, Any]) -> int:
        deleted = 0
        docs = await self.list_documents(filter=filter, limit=100000)
        ids = [d.get("id") for d in docs if d.get("id") is not None]
        if not ids:
            return 0
        if hasattr(self.client, "delete"):
            try:
                await self.client.delete(self.collection, points_selector={"points": ids})
            except TypeError:
                await self.client.delete(self.collection, ids=ids)
        deleted = len(ids)
        return deleted

    async def update_documents(self, filter: Dict[str, Any], update_data: Dict[str, Any]) -> int:
        if not update_data:
            return 0
        docs = await self.list_documents(filter=filter, limit=100000)
        ids = [d.get("id") for d in docs if d.get("id") is not None]
        if not ids:
            return 0
        payload: Dict[str, Any] = {}
        if "content" in update_data and update_data["content"] is not None:
            payload["content"] = str(update_data["content"])
        if "metadata" in update_data and isinstance(update_data["metadata"], dict):
            payload["metadata"] = update_data["metadata"]
        if not payload:
            return 0
        if hasattr(self.client, "set_payload"):
            try:
                await self.client.set_payload(self.collection, payload=payload, points=ids)
            except TypeError:
                await self.client.set_payload(collection_name=self.collection, payload=payload, points=ids)
            return len(ids)
        if hasattr(self.client, "upsert"):
            for d in docs:
                pid = d.get("id")
                if pid is None:
                    continue
                merged_md = d.get("metadata") or {}
                if "metadata" in payload and isinstance(payload["metadata"], dict):
                    merged_md = payload["metadata"]
                merged_content = payload.get("content", d.get("content", ""))
                await self.client.upsert(
                    self.collection,
                    points=[{"id": pid, "payload": {"content": merged_content, "metadata": merged_md}}],
                )
            return len(ids)
        raise NotImplementedError("Qdrant client does not support payload update")



from typing import List, Dict, Any, Optional
import json

class InMemoryHistory:
    def __init__(self, max_messages: int = 20):
        self.sessions: Dict[str, List[Dict[str, Any]]] = {}
        self.max_messages = max_messages
    def add_message(self, session_id: str, role: str, content: str):
        if not session_id:
            return
        arr = self.sessions.get(session_id, [])
        arr.append({ 'role': role, 'content': content })
        start = max(0, len(arr) - self.max_messages)
        self.sessions[session_id] = arr[start:]
    def get_recent(self, session_id: str, n: int = 10) -> List[Dict[str, Any]]:
        arr = self.sessions.get(session_id, [])
        start = max(0, len(arr) - n)
        return arr[start:]

class RedisHistory:
    def __init__(self, client: Any, key_prefix: str = 'vectra:chat:', max_messages: int = 20):
        self.client = client
        self.key_prefix = key_prefix
        self.max_messages = max_messages
    async def add_message(self, session_id: str, role: str, content: str):
        if not session_id or not self.client:
            return
        key = f"{self.key_prefix}{session_id}"
        payload = json.dumps({ 'role': role, 'content': content })
        try:
            if hasattr(self.client, 'rpush'):
                await self.client.rpush(key, payload)
            elif hasattr(self.client, 'lpush'):
                await self.client.lpush(key, payload)
            if hasattr(self.client, 'ltrim'):
                await self.client.ltrim(key, -self.max_messages, -1)
        except Exception:
            pass
    async def get_recent(self, session_id: str, n: int = 10) -> List[Dict[str, Any]]:
        if not session_id or not self.client:
            return []
        key = f"{self.key_prefix}{session_id}"
        try:
            arr = []
            if hasattr(self.client, 'lrange'):
                arr = await self.client.lrange(key, -n, -1)
            elif hasattr(self.client, 'lRange'):
                arr = await self.client.lRange(key, -n, -1)
            out: List[Dict[str, Any]] = []
            for x in arr:
                try:
                    out.append(json.loads(x))
                except Exception:
                    out.append({ 'role': 'assistant', 'content': str(x) })
            return out
        except Exception:
            return []

class PostgresHistory:
    def __init__(self, client: Any, table_name: str = 'ChatMessage', column_map: Optional[Dict[str, str]] = None, max_messages: int = 20):
        self.client = client
        self.table_name = table_name
        self.column_map = column_map or { 'sessionId': 'sessionId', 'role': 'role', 'content': 'content', 'createdAt': 'createdAt' }
        self.max_messages = max_messages
    async def add_message(self, session_id: str, role: str, content: str):
        if not session_id or not self.client:
            return
        t = self.table_name
        c = self.column_map
        q = f'INSERT INTO "{t}" ("{c["sessionId"]}","{c["role"]}","{c["content"]}","{c["createdAt"]}") VALUES ($1,$2,$3,NOW())'
        try:
            if hasattr(self.client, 'execute_raw'):
                await self.client.execute_raw(q, session_id, role, content)
        except Exception:
            pass
    async def get_recent(self, session_id: str, n: int = 10) -> List[Dict[str, Any]]:
        if not session_id or not self.client:
            return []
        t = self.table_name
        c = self.column_map
        q = f'SELECT "{c["role"]}" as role, "{c["content"]}" as content FROM "{t}" WHERE "{c["sessionId"]}" = $1 ORDER BY "{c["createdAt"]}" DESC LIMIT {max(1,n)}'
        try:
            rows = []
            if hasattr(self.client, 'query_raw'):
                rows = await self.client.query_raw(q, session_id)
            if isinstance(rows, list):
                rows.reverse()
                return [{ 'role': r.get('role'), 'content': r.get('content') } for r in rows]
            return []
        except Exception:
            return []



from typing import List, Dict, Any, Optional
from ..interfaces import VectorStore

class MilvusVectorStore(VectorStore):
    def __init__(self, config):
        self.config = config
        self.client = config.client_instance
        self.collection = config.table_name or 'rag_collection'

    async def add_documents(self, documents: List[Dict[str, Any]]):
        data = [{ 'vector': d['embedding'], 'content': d['content'], 'metadata': d['metadata'] } for d in documents]
        await self.client.insert(collection_name=self.collection, fields_data=data)

    async def similarity_search(self, vector: List[float], limit: int = 5, filter: Optional[Dict] = None) -> List[Dict[str, Any]]:
        expr = self._filter_to_expr(filter)
        if expr:
            try:
                res = await self.client.search(collection_name=self.collection, data=[vector], limit=limit, expr=expr)
            except TypeError:
                res = await self.client.search(collection_name=self.collection, data=[vector], limit=limit, filter=expr)
        else:
            res = await self.client.search(collection_name=self.collection, data=[vector], limit=limit)
        hits = res.get('results', []) if isinstance(res, dict) else res
        return [{ 'content': h.get('content', ''), 'metadata': h.get('metadata', {}), 'score': h.get('distance', 0.0) } for h in hits]

    async def hybrid_search(self, text: str, vector: List[float], limit: int = 5, filter: Optional[Dict] = None) -> List[Dict[str, Any]]:
        return await self.similarity_search(vector, limit, filter)

    def _filter_to_expr(self, filter: Optional[Dict[str, Any]]) -> str:
        if not filter:
            return ""
        parts: List[str] = []
        for k, v in filter.items():
            if isinstance(v, str):
                parts.append(f'metadata["{k}"] == "{v}"')
            elif isinstance(v, bool):
                parts.append(f'metadata["{k}"] == {str(v).lower()}')
            elif isinstance(v, (int, float)):
                parts.append(f'metadata["{k}"] == {v}')
        return " and ".join(parts)

    async def list_documents(self, filter: Optional[Dict[str, Any]] = None, limit: int = 100) -> List[Dict[str, Any]]:
        expr = self._filter_to_expr(filter)
        if not hasattr(self.client, "query"):
            raise NotImplementedError("Milvus client does not support query()")
        res = await self.client.query(
            collection_name=self.collection,
            expr=expr,
            output_fields=["content", "metadata"],
            limit=max(1, int(limit)),
        )
        out: List[Dict[str, Any]] = []
        for r in res or []:
            out.append({"content": r.get("content", ""), "metadata": r.get("metadata") or {}})
        return out

    async def delete_documents(self, filter: Dict[str, Any]) -> int:
        expr = self._filter_to_expr(filter)
        if not hasattr(self.client, "delete"):
            raise NotImplementedError("Milvus client does not support delete()")
        await self.client.delete(collection_name=self.collection, expr=expr)
        return 0

    async def update_documents(self, filter: Dict[str, Any], update_data: Dict[str, Any]) -> int:
        raise NotImplementedError("Milvus update_documents is not implemented")



class LoggingCallbackHandler:
    def on_ingest_start(self, file_path): print(f"[RAG] Starting ingestion: {file_path}")
    def on_ingest_end(self, file_path, chunk_count, duration_ms=None): print(f"[RAG] Finished ingestion. Chunks: {chunk_count} ({duration_ms} ms)")
    def on_ingest_summary(self, summary): print(f"[RAG] Ingest summary: processed={summary['processed']}, ok={summary['succeeded']}, failed={summary['failed']}")
    def on_chunking_start(self, strategy): print(f"[RAG] Chunking strategy: {strategy}")
    def on_embedding_start(self, count): print(f"[RAG] Embedding {count} chunks...")
    def on_retrieval_start(self, query): print(f"[RAG] Querying: \"{query}\"")
    def on_retrieval_end(self, count, duration_ms=None): print(f"[RAG] Retrieved {count} docs ({duration_ms} ms)")
    def on_reranking_start(self, count): print(f"[RAG] Reranking {count} docs...")
    def on_reranking_end(self, count): print(f"[RAG] Reranking finished. Keeping top {count}.")
    def on_generation_start(self, prompt_preview=None): print("[RAG] Generating answer...")
    def on_generation_end(self, answer_preview=None, duration_ms=None): print(f"[RAG] Answer generated ({duration_ms} ms)")
    def on_error(self, err): print(f"[RAG] Error: {getattr(err, 'message', str(err))}")

class StructuredLoggingCallbackHandler:
    def on_ingest_start(self, file_path): print({ 'event': 'ingest_start', 'filePath': file_path })
    def on_ingest_end(self, file_path, chunk_count, duration_ms=None): print({ 'event': 'ingest_end', 'filePath': file_path, 'chunkCount': chunk_count, 'durationMs': duration_ms })
    def on_ingest_summary(self, summary): print({ 'event': 'ingest_summary', **summary })
    def on_chunking_start(self, strategy): print({ 'event': 'chunking_start', 'strategy': strategy })
    def on_embedding_start(self, count): print({ 'event': 'embedding_start', 'count': count })
    def on_retrieval_start(self, query): print({ 'event': 'retrieval_start', 'query': query })
    def on_retrieval_end(self, count, duration_ms=None): print({ 'event': 'retrieval_end', 'count': count, 'durationMs': duration_ms })
    def on_reranking_start(self, count): print({ 'event': 'reranking_start', 'count': count })
    def on_reranking_end(self, count): print({ 'event': 'reranking_end', 'count': count })
    def on_generation_start(self, prompt_preview=None): print({ 'event': 'generation_start', 'promptPreview': str(prompt_preview)[:120] if prompt_preview else None })
    def on_generation_end(self, answer_preview=None, duration_ms=None): print({ 'event': 'generation_end', 'answerPreview': str(answer_preview)[:120] if answer_preview else None, 'durationMs': duration_ms })
    def on_error(self, err): print({ 'event': 'error', 'message': getattr(err, 'message', str(err)) })



import os
import asyncio
from typing import List, AsyncGenerator
from openai import AsyncOpenAI
from ..config import EmbeddingConfig, LLMConfig

class OpenAIBackend:
    def __init__(self, config: LLMConfig | EmbeddingConfig):
        self.config = config
        api_key = config.api_key or os.getenv('OPENAI_API_KEY')
        if not api_key:
            raise ValueError("OpenAI API Key missing.")
        self.client = AsyncOpenAI(api_key=api_key)

    async def embed_documents(self, texts: List[str]) -> List[List[float]]:
        res = await self.client.embeddings.create(
            input=texts,
            model=self.config.model_name,
            dimensions=self.config.dimensions or 1536
        )
        return [d.embedding for d in res.data]

    async def embed_query(self, text: str) -> List[float]:
        res = await self.client.embeddings.create(
            input=[text],
            model=self.config.model_name,
            dimensions=self.config.dimensions
        )
        return res.data[0].embedding

    async def generate(self, prompt: str, system_instruction: str = "You are a helpful assistant.") -> str:
        res = await self.client.chat.completions.create(
            model=self.config.model_name,
            messages=[
                {"role": "system", "content": system_instruction},
                {"role": "user", "content": prompt}
            ],
            temperature=self.config.temperature,
            max_tokens=self.config.max_tokens
        )
        return res.choices[0].message.content

    async def generate_stream(self, prompt: str, system_instruction: str = "You are a helpful assistant.") -> AsyncGenerator[dict, None]:
        stream = await self.client.chat.completions.create(
            model=self.config.model_name,
            messages=[
                {"role": "system", "content": system_instruction},
                {"role": "user", "content": prompt}
            ],
            temperature=self.config.temperature,
            max_tokens=self.config.max_tokens,
            stream=True
        )
        async for chunk in stream:
            if chunk.choices[0].delta.content:
                yield { 'delta': chunk.choices[0].delta.content, 'finish_reason': None, 'usage': None }



import json
import asyncio
import requests

class OllamaBackend:
    def __init__(self, config):
        self.config = config
        self.base_url = getattr(config, 'base_url', None) or 'http://localhost:11434'
    async def embed_documents(self, texts):
        out = []
        for t in texts:
            r = requests.post(f"{self.base_url}/api/embeddings", json={ 'model': self.config.model_name, 'prompt': t })
            j = r.json()
            out.append(j.get('embedding') or j.get('data') or [])
        return out
    async def embed_query(self, text):
        r = requests.post(f"{self.base_url}/api/embeddings", json={ 'model': self.config.model_name, 'prompt': text })
        j = r.json()
        return j.get('embedding') or j.get('data') or []
    async def generate(self, prompt, sys=None):
        r = requests.post(f"{self.base_url}/api/generate", json={ 'model': self.config.model_name, 'prompt': (f"{sys}\n\n{prompt}" if sys else prompt), 'stream': False })
        j = r.json()
        return j.get('response','')
    async def generate_stream(self, prompt, sys=None):
        text = await self.generate(prompt, sys)
        for i in range(0, len(text), 64):
            yield { 'delta': text[i:i+64], 'finish_reason': None, 'usage': None }



from .core import VectraClient
from .config import (
    VectraConfig,
    EmbeddingConfig,
    LLMConfig,
    ChunkingConfig,
    RerankingConfig,
    RetrievalConfig,
    DatabaseConfig,
    ProviderType,
    ChunkingStrategy,
    RetrievalStrategy
)

__all__ = [
    'VectraClient',
    'VectraConfig',
    'EmbeddingConfig',
    'LLMConfig',
    'ChunkingConfig',
    'RerankingConfig',
    'RetrievalConfig',
    'DatabaseConfig',
    'ProviderType',
    'ChunkingStrategy',
    'RetrievalStrategy'
]



from typing import List, Dict, Any

async def evaluate_test_set(client, test_set: List[Dict[str, str]]) -> List[Dict[str, Any]]:
    report: List[Dict[str, Any]] = []
    for item in test_set:
        res = await client.query_rag(item['question'])
        ctx = "\n".join([s.get('summary','') for s in res.get('sources', [])])
        fp = f"Rate 0-1: Is the following Answer derived only from the Context?\nContext:\n{ctx}\n\nAnswer:\n{res.get('answer') if isinstance(res.get('answer'), str) else str(res.get('answer'))}"
        rp = f"Rate 0-1: Does the Answer correctly answer the Question?\nQuestion:\n{item['question']}\n\nAnswer:\n{res.get('answer') if isinstance(res.get('answer'), str) else str(res.get('answer'))}"
        faith = 0.0; rel = 0.0
        try:
            faith = max(0.0, min(1.0, float(str(await client.llm.generate(fp, 'You return a single number between 0 and 1.')))))
        except Exception:
            faith = 0.0
        try:
            rel = max(0.0, min(1.0, float(str(await client.llm.generate(rp, 'You return a single number between 0 and 1.')))))
        except Exception:
            rel = 0.0
        report.append({ 'question': item['question'], 'expectedGroundTruth': item.get('expectedGroundTruth',''), 'faithfulness': faith, 'relevance': rel })
    return report



import os
import asyncio
from typing import AsyncGenerator
from anthropic import AsyncAnthropic
from ..config import LLMConfig

class AnthropicBackend:
    def __init__(self, config: LLMConfig):
        self.config = config
        api_key = config.api_key or os.getenv('ANTHROPIC_API_KEY')
        if not api_key:
            raise ValueError("Anthropic API Key missing.")
        self.client = AsyncAnthropic(api_key=api_key)

    async def embed_documents(self, texts):
        raise NotImplementedError("Anthropic does not provide embedding models.")

    async def embed_query(self, text):
        raise NotImplementedError("Anthropic does not provide embedding models.")

    async def generate(self, prompt: str, system_instruction: str = "You are a helpful assistant.") -> str:
        res = await self.client.messages.create(
            model=self.config.model_name,
            max_tokens=self.config.max_tokens,
            temperature=self.config.temperature,
            system=system_instruction,
            messages=[{"role": "user", "content": prompt}]
        )
        return res.content[0].text
    
    async def generate_stream(self, prompt: str, system_instruction: str = "You are a helpful assistant.") -> AsyncGenerator[dict, None]:
        stream = await self.client.messages.create(
            model=self.config.model_name,
            max_tokens=self.config.max_tokens,
            temperature=self.config.temperature,
            system=system_instruction,
            messages=[{"role": "user", "content": prompt}],
            stream=True
        )
        async for chunk in stream:
            if chunk.type == 'content_block_delta':
                yield { 'delta': chunk.delta.text, 'finish_reason': None, 'usage': None }



import uuid
from typing import List, Dict, Any, Optional
from ..interfaces import VectorStore
from ..config import DatabaseConfig

class ChromaVectorStore(VectorStore):
    def __init__(self, config: DatabaseConfig):
        # Expecting 'client_instance' to be a chromadb.Client or PersistentClient
        self.client = config.client_instance
        self.collection_name = config.table_name or "rag_collection"
        self.collection = self.client.get_or_create_collection(name=self.collection_name)

    async def add_documents(self, documents: List[Dict[str, Any]]):
        ids = [str(uuid.uuid4()) for _ in documents]
        embeddings = [doc['embedding'] for doc in documents]
        metadatas = [doc['metadata'] for doc in documents]
        documents_text = [doc['content'] for doc in documents]
        
        self.collection.add(
            ids=ids,
            embeddings=embeddings,
            metadatas=metadatas,
            documents=documents_text
        )
    
    async def file_exists(self, sha256: str, size: int, last_modified: int) -> bool:
        try:
            res = self.collection.get(where={"fileSHA256": sha256, "fileSize": size, "lastModified": last_modified})
            return bool(res and res.get('ids'))
        except Exception:
            return False

    async def similarity_search(self, vector: List[float], limit: int = 5, filter: Optional[Dict] = None) -> List[Dict[str, Any]]:
        results = self.collection.query(
            query_embeddings=[vector],
            n_results=limit,
            where=filter
        )
        
        # Unpack Chroma structure
        # results['documents'][0] is list of docs for first query
        if not results['documents']: return []
        
        out = []
        for i in range(len(results['documents'][0])):
            out.append({
                'content': results['documents'][0][i],
                'metadata': results['metadatas'][0][i],
                'score': 1.0 - (results['distances'][0][i] if 'distances' in results else 0) 
            })
        return out

    async def hybrid_search(self, text: str, vector: List[float], limit: int = 5, filter: Optional[Dict] = None) -> List[Dict[str, Any]]:
        return await self.similarity_search(vector, limit, filter)

    async def list_documents(self, filter: Optional[Dict[str, Any]] = None, limit: int = 100) -> List[Dict[str, Any]]:
        kwargs: Dict[str, Any] = {}
        if filter:
            kwargs["where"] = filter
        limit_int = max(1, int(limit))
        try:
            res = self.collection.get(limit=limit_int, **kwargs)
        except TypeError:
            res = self.collection.get(**kwargs)
        ids = res.get("ids") or []
        documents = res.get("documents") or []
        metadatas = res.get("metadatas") or []
        out: List[Dict[str, Any]] = []
        for i in range(min(len(ids), len(documents))):
            out.append(
                {
                    "id": ids[i],
                    "content": documents[i],
                    "metadata": metadatas[i] if i < len(metadatas) else {},
                }
            )
        return out[:limit_int]

    async def delete_documents(self, filter: Dict[str, Any]) -> int:
        deleted = 0
        offset = 0
        batch = 1000
        while True:
            try:
                res = self.collection.get(where=filter, limit=batch, offset=offset)
            except TypeError:
                res = self.collection.get(where=filter)
            ids = res.get("ids") or []
            if not ids:
                break
            self.collection.delete(ids=ids)
            deleted += len(ids)
            if len(ids) < batch:
                break
            offset += batch
        return deleted

    async def update_documents(self, filter: Dict[str, Any], update_data: Dict[str, Any]) -> int:
        if not update_data:
            return 0
        docs = await self.list_documents(filter=filter, limit=100000)
        if not docs:
            return 0
        new_content = update_data.get("content", None)
        update_meta = update_data.get("metadata", None)
        ids: List[str] = []
        documents: List[str] = []
        metadatas: List[Dict[str, Any]] = []
        for d in docs:
            ids.append(d["id"])
            documents.append(new_content if isinstance(new_content, str) else d.get("content", ""))
            md = d.get("metadata") or {}
            if isinstance(update_meta, dict):
                merged = dict(md)
                merged.update(update_meta)
                metadatas.append(merged)
            else:
                metadatas.append(md)
        self.collection.update(ids=ids, documents=documents, metadatas=metadatas)
        return len(ids)



from abc import ABC, abstractmethod
from typing import List, Dict, Any, Optional

class VectorStore(ABC):
    @abstractmethod
    async def add_documents(self, documents: List[Dict[str, Any]]):
        pass

    @abstractmethod
    async def similarity_search(self, vector: List[float], limit: int = 5, filter: Optional[Dict] = None) -> List[Dict[str, Any]]:
        pass
    
    # New method for Hybrid Search support
    @abstractmethod
    async def hybrid_search(self, text: str, vector: List[float], limit: int = 5, filter: Optional[Dict] = None) -> List[Dict[str, Any]]:
        """
        Perform hybrid search (Sparse + Dense). 
        Default implementation falls back to similarity_search if not supported.
        """
        return await self.similarity_search(vector, limit, filter)
    
    async def file_exists(self, sha256: str, size: int, last_modified: int) -> bool:
        return False

    @abstractmethod
    async def delete_documents(self, filter: Dict[str, Any]) -> int:
        raise NotImplementedError

    @abstractmethod
    async def update_documents(self, filter: Dict[str, Any], update_data: Dict[str, Any]) -> int:
        raise NotImplementedError

    @abstractmethod
    async def list_documents(self, filter: Optional[Dict[str, Any]] = None, limit: int = 100) -> List[Dict[str, Any]]:
        raise NotImplementedError



import json
import logging
import uuid
from typing import List, Dict, Any, Optional
import asyncio
from ..interfaces import VectorStore

def to_db_vector(vector: List[float]) -> str:
    return f"[{','.join(map(str, vector))}]"

class PostgresVectorStore(VectorStore):
    def __init__(self, config: Any):
        self.config = config
        self.client = config.client_instance
        self.table_name = getattr(config, 'table_name', 'document')
        self.column_map = getattr(config, 'column_map', {})
        self.c_content = self.column_map.get('content', 'content')
        self.c_meta = self.column_map.get('metadata', 'metadata')
        self.c_vector = self.column_map.get('vector', 'vector')
        
    async def ensure_indexes(self, dimensions: int = 1536):
        await self.client.execute("CREATE EXTENSION IF NOT EXISTS vector")
        
        # Default dimension 1536 (OpenAI/Gemini-ish), but user should manage schema for other dims
        create_table_sql = f"""
        CREATE TABLE IF NOT EXISTS "{self.table_name}" (
            "id" TEXT PRIMARY KEY,
            "{self.c_content}" TEXT,
            "{self.c_meta}" JSONB,
            "{self.c_vector}" vector({dimensions}),
            "createdAt" TIMESTAMP WITH TIME ZONE DEFAULT NOW()
        );
        """
        await self.client.execute(create_table_sql)
        
        index_sql = f"""
        CREATE INDEX IF NOT EXISTS "{self.table_name}_vec_idx" 
        ON "{self.table_name}" USING hnsw ("{self.c_vector}" vector_cosine_ops);
        """
        await self.client.execute(index_sql)

    async def add_documents(self, documents: List[Dict[str, Any]]):
        sql = f"""
        INSERT INTO "{self.table_name}" ("id", "{self.c_content}", "{self.c_meta}", "{self.c_vector}", "createdAt")
        VALUES ($1, $2, $3, $4, NOW())
        ON CONFLICT ("id") DO UPDATE SET
            "{self.c_content}" = EXCLUDED."{self.c_content}",
            "{self.c_meta}" = EXCLUDED."{self.c_meta}",
            "{self.c_vector}" = EXCLUDED."{self.c_vector}",
            "createdAt" = NOW();
        """
        data = []
        for doc in documents:
            doc_id = doc.get('id') or str(uuid.uuid4())
            content = doc.get('content')
            meta = json.dumps(doc.get('metadata', {}))
            vec = to_db_vector(doc.get('embedding', []))
            data.append((doc_id, content, meta, vec))
            
        await self.client.executemany(sql, data)

    async def similarity_search(self, vector: List[float], limit: int = 5, filter: Optional[Dict] = None) -> List[Dict[str, Any]]:
        vec_str = to_db_vector(vector)
        where_clause = ""
        params = [vec_str, limit]
        
        if filter:
            where_clause = f'WHERE "{self.c_meta}" @> $3::jsonb'
            params.append(json.dumps(filter))

        sql = f"""
        SELECT "id", "{self.c_content}" as content, "{self.c_meta}" as metadata, 
               ("{self.c_vector}" <=> $1) as distance
        FROM "{self.table_name}"
        {where_clause}
        ORDER BY distance ASC
        LIMIT $2
        """
        
        rows = await self.client.fetch(sql, *params)
        
        results = []
        for row in rows:
            results.append({
                'id': row['id'],
                'content': row['content'],
                'metadata': json.loads(row['metadata']) if isinstance(row['metadata'], str) else row['metadata'],
                'score': 1 - row['distance']
            })
        return results

    async def hybrid_search(self, text: str, vector: List[float], limit: int = 5, filter: Optional[Dict] = None) -> List[Dict[str, Any]]:
        # Fallback to similarity search for now
        return await self.similarity_search(vector, limit, filter)

    async def delete_documents(self, filter: Dict[str, Any]) -> int:
        sql = f'DELETE FROM "{self.table_name}" WHERE "{self.c_meta}" @> $1::jsonb'
        res = await self.client.execute(sql, json.dumps(filter))
        return int(res.split(' ')[1]) if res else 0
        
    async def update_documents(self, filter: Dict[str, Any], update_data: Dict[str, Any]) -> int:
        raise NotImplementedError

    async def list_documents(self, filter: Optional[Dict[str, Any]] = None, limit: int = 100) -> List[Dict[str, Any]]:
        where_clause = ""
        params = [limit]
        if filter:
            where_clause = f'WHERE "{self.c_meta}" @> $2::jsonb'
            params.append(json.dumps(filter))
            
        sql = f"""
        SELECT "id", "{self.c_content}" as content, "{self.c_meta}" as metadata
        FROM "{self.table_name}"
        {where_clause}
        LIMIT $1
        """
        rows = await self.client.fetch(sql, *params)
        return [{
            'id': r['id'],
            'content': r['content'],
            'metadata': json.loads(r['metadata']) if isinstance(r['metadata'], str) else r['metadata']
        } for r in rows]



import json
import os
import threading
import webbrowser
import sqlite3
from urllib.parse import urlparse, parse_qs
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from .config import VectraConfig, ProviderType, ChunkingStrategy, RetrievalStrategy

def _get_db_connection(config_path):
    try:
        if not os.path.exists(config_path):
            return None
        with open(config_path, 'r', encoding='utf-8') as f:
            cfg = json.load(f)
        
        obs = cfg.get('observability', {})
        if not obs.get('enabled'):
            return None
            
        db_path = obs.get('sqlite_path', 'vectra-observability.db')
        conn = sqlite3.connect(db_path)
        conn.row_factory = sqlite3.Row
        return conn
    except Exception as e:
        print(f"Failed to open observability DB: {e}")
        return None

def _default_config():
    return {
        "embedding": {
            "provider": ProviderType.OPENAI.value,
            "api_key": "",
            "model_name": "text-embedding-3-small",
            "dimensions": None
        },
        "llm": {
            "provider": ProviderType.GEMINI.value,
            "api_key": "",
            "model_name": "gemini-1.5-pro-latest",
            "temperature": 0.0,
            "max_tokens": 1024,
            "base_url": None,
            "default_headers": None
        },
        "database": {
            "type": "prisma",
            "table_name": "Document",
            "column_map": {"content": "content", "vector": "vector", "metadata": "metadata"}
        },
        "chunking": {
            "strategy": ChunkingStrategy.RECURSIVE.value,
            "chunk_size": 1000,
            "chunk_overlap": 200,
            "separators": ["\n\n", "\n", " ", ""]
        },
        "retrieval": {
            "strategy": RetrievalStrategy.NAIVE.value,
            "hybrid_alpha": 0.5
        },
        "reranking": {
            "enabled": False,
            "provider": "llm",
            "top_n": 5,
            "window_size": 20
        },
        "metadata": None,
        "query_planning": None,
        "grounding": None,
        "generation": None,
        "prompts": None,
        "tracing": None,
        "callbacks": []
    }

class _Handler(BaseHTTPRequestHandler):
    def _send_json(self, status, obj):
        data = json.dumps(obj, ensure_ascii=False).encode("utf-8")
        self.send_response(status)
        self.send_header("Content-Type", "application/json")
        self.send_header("Content-Length", str(len(data)))
        self.end_headers()
        self.wfile.write(data)

    def _serve_static(self, path, content_type, folder='ui'):
        try:
            # Locate the ui directory relative to this file
            base_dir = os.path.dirname(__file__)
            file_path = os.path.join(base_dir, folder, path)
            
            with open(file_path, 'rb') as f:
                data = f.read()
                
            self.send_response(200)
            self.send_header("Content-Type", content_type)
            self.send_header("Content-Length", str(len(data)))
            self.end_headers()
            self.wfile.write(data)
        except Exception:
            self.send_error(404)

    def do_GET(self):
        p = urlparse(self.path)
        
        # --- Dashboard Routes ---
        if p.path == "/dashboard":
            self.send_response(301)
            self.send_header('Location', '/dashboard/')
            self.end_headers()
            return
            
        if p.path == "/dashboard/":
            self._serve_static("index.html", "text/html; charset=utf-8", folder="dashboard")
            return
            
        if p.path.startswith("/dashboard/"):
            asset_name = p.path.replace("/dashboard/", "")
            content_type = "text/plain"
            if asset_name.endswith(".css"): content_type = "text/css"
            elif asset_name.endswith(".js"): content_type = "application/javascript"
            elif asset_name.endswith(".html"): content_type = "text/html"
            elif asset_name.endswith(".png"): content_type = "image/png"
            self._serve_static(asset_name, content_type, folder="dashboard")
            return

        # --- Observability API ---
        if p.path.startswith("/api/observability/"):
            print(f"DEBUG: Handling observability request: {p.path}")
            conn = _get_db_connection(self.server.config_path)
            if not conn:
                self._send_json(400, {"error": "Observability not enabled or DB not found"})
                return

            try:
                qs = parse_qs(p.query)
                project_id = qs.get('projectId', [None])[0]
                
                # Filter Logic
                where_clauses = []
                params = []
                if project_id and project_id != 'all':
                    where_clauses.append("project_id = ?")
                    params.append(project_id)
                
                where_sql = "WHERE " + " AND ".join(where_clauses) if where_clauses else ""
                and_sql = "AND " + " AND ".join(where_clauses) if where_clauses else ""

                if p.path.endswith("/stats"):
                    cur = conn.cursor()
                    
                    # Total Requests
                    cur.execute(f"SELECT COUNT(*) as count FROM traces WHERE name = 'query_rag' {and_sql.replace('AND', 'WHERE', 1) if not where_sql else and_sql}", params)
                    total_req = cur.fetchone()['count']
                    
                    # Avg Latency
                    cur.execute(f"SELECT AVG(value) as val FROM metrics WHERE name = 'query_latency' {and_sql}", params)
                    avg_lat = cur.fetchone()['val'] or 0
                    
                    # Token Counts
                    cur.execute(f"SELECT SUM(value) as val FROM metrics WHERE name = 'prompt_chars' {and_sql}", params)
                    tokens_p = cur.fetchone()['val'] or 0
                    
                    # Error Rate
                    cur.execute(f"SELECT COUNT(*) as count FROM traces WHERE name = 'query_rag' AND status = 'error' {and_sql}", params)
                    error_count = cur.fetchone()['count']
                    error_rate = (error_count / total_req * 100) if total_req > 0 else 0
                    
                    cur.execute(f"SELECT SUM(value) as val FROM metrics WHERE name = 'completion_chars' {and_sql}", params)
                    tokens_c = cur.fetchone()['val'] or 0

                    # History
                    cur.execute(f"""
                        SELECT m.timestamp, m.value as latency, 
                        (SELECT value FROM metrics m2 WHERE m2.timestamp = m.timestamp AND m2.name = 'prompt_chars') + 
                        (SELECT value FROM metrics m3 WHERE m3.timestamp = m.timestamp AND m3.name = 'completion_chars') as tokens
                        FROM metrics m 
                        WHERE m.name = 'query_latency' {and_sql}
                        ORDER BY m.timestamp DESC LIMIT 50
                    """, params)
                    history = [dict(row) for row in cur.fetchall()]
                    history.reverse()

                    self._send_json(200, {
                        "totalRequests": total_req,
                        "avgLatency": avg_lat,
                        "totalPromptChars": tokens_p,
                        "totalCompletionChars": tokens_c,
                        "errorRate": error_rate,
                        "history": history
                    })
                
                elif p.path.endswith("/projects"):
                    cur = conn.cursor()
                    cur.execute("SELECT DISTINCT project_id FROM traces")
                    rows = cur.fetchall()
                    projects = [row['project_id'] for row in rows if row['project_id']]
                    self._send_json(200, projects)

                elif p.path.endswith("/traces"):
                    cur = conn.cursor()
                    cur.execute("SELECT * FROM traces WHERE name = 'query_rag' ORDER BY start_time DESC LIMIT 50")
                    rows = [dict(row) for row in cur.fetchall()]
                    self._send_json(200, rows)

                elif "/traces/" in p.path:
                    trace_id = p.path.split("/")[-1]
                    cur = conn.cursor()
                    cur.execute("SELECT * FROM traces WHERE trace_id = ?", (trace_id,))
                    rows = [dict(row) for row in cur.fetchall()]
                    self._send_json(200, rows)

                elif p.path.endswith("/sessions"):
                    cur = conn.cursor()
                    # Fix session filter
                    s_where = where_sql.replace('project_id', 'project_id') 
                    cur.execute(f"SELECT * FROM sessions {s_where} ORDER BY last_active DESC LIMIT 50", params)
                    rows = [dict(row) for row in cur.fetchall()]
                    
                    # Parse metadata and keep snake_case to match JS dashboard
                    results = []
                    for r in rows:
                        d = dict(r)
                        if d.get('metadata'):
                            try:
                                d['metadata'] = json.loads(d['metadata'])
                            except:
                                d['metadata'] = {}
                        else:
                            d['metadata'] = {}
                        results.append(d)
                        
                    self._send_json(200, results)
                
                else:
                    self._send_json(404, {"error": "Unknown endpoint"})

            except Exception as e:
                self._send_json(500, {"error": str(e)})
            finally:
                conn.close()
            return

        # --- Legacy Config UI ---
        if p.path == "/" or p.path == "/index.html":
            self._serve_static("index.html", "text/html; charset=utf-8")
            return
            
        if p.path == "/style.css":
            self._serve_static("style.css", "text/css")
            return
            
        if p.path == "/script.js":
            self._serve_static("script.js", "application/javascript")
            return
            
        if p.path == "/config":
            cfg_path = self.server.config_path
            if os.path.exists(cfg_path):
                try:
                    with open(cfg_path, "r", encoding="utf-8") as f:
                        raw = json.load(f)
                    val = VectraConfig.model_validate(raw)
                    self._send_json(200, val.model_dump())
                    return
                except Exception as e:
                    self._send_json(400, {"error": str(e)})
                    return
            self._send_json(200, _default_config())
            return
        self.send_error(404)

    def do_POST(self):
        p = urlparse(self.path)
        if p.path == "/config":
            ln = int(self.headers.get("Content-Length") or 0)
            body = self.rfile.read(ln) if ln > 0 else b"{}"
            try:
                raw = json.loads(body.decode("utf-8"))
                val = VectraConfig.model_validate(raw)
                out = val.model_dump(exclude_none=True)
                os.makedirs(os.path.dirname(self.server.config_path) or ".", exist_ok=True)
                with open(self.server.config_path, "w", encoding="utf-8") as f:
                    json.dump(out, f, ensure_ascii=False, indent=2)
                self._send_json(200, {"message": "Saved"})
            except Exception as e:
                self._send_json(400, {"error": str(e)})
            return
        self.send_error(404)

def start(config_path, mode='webconfig', host="127.0.0.1", port=8765, open_browser=True):
    server = None
    while server is None:
        try:
            server = ThreadingHTTPServer((host, port), _Handler)
        except OSError:
            print(f"Port {port} is in use, trying {port + 1}...")
            port += 1
            if port > 65535:
                raise Exception("No available ports found")

    server.config_path = os.path.abspath(config_path)
    t = threading.Thread(target=server.serve_forever, daemon=True)
    t.start()
    url = f"http://{host}:{port}/"
    if mode == 'dashboard':
        url = f"http://{host}:{port}/dashboard"
    print(f"Vectra WebConfig running at {url}")
    if open_browser:
        try:
            webbrowser.open(url)
        except Exception:
            pass
    return server




import os
import asyncio
from typing import List, AsyncGenerator
from google import genai
from google.genai import types
from ..config import EmbeddingConfig, LLMConfig

import logging

logger = logging.getLogger(__name__)

class GeminiBackend:
    def __init__(self, config: LLMConfig | EmbeddingConfig):
        self.config = config
        api_key = config.api_key or os.getenv('GOOGLE_API_KEY')
        if not api_key:
            raise ValueError("Gemini API Key missing.")
        self.client = genai.Client(api_key=api_key)

    async def _retry(self, fn, retries=3):
        for i in range(retries):
            try:
                return await fn()
            except Exception as e:
                logger.error(f"Gemini API error (attempt {i+1}/{retries}): {e}")
                if i == retries - 1: raise e
                await asyncio.sleep(1 * (2 ** i))

    async def embed_documents(self, texts: List[str]) -> List[List[float]]:
        embeddings = []
        BATCH_SIZE = 100
        logger.info(f"Embedding {len(texts)} documents with model {self.config.model_name}")
        for i in range(0, len(texts), BATCH_SIZE):
            batch = texts[i:i + BATCH_SIZE]
            
            async def _call_embed():
                 response = await self.client.aio.models.embed_content(
                    model=self.config.model_name,
                    contents=batch,
                    config=types.EmbedContentConfig(
                         task_type="RETRIEVAL_DOCUMENT",
                         output_dimensionality=self.config.dimensions if getattr(self.config, 'dimensions', None) else None
                    )
                )
                 return response

            res = await self._retry(_call_embed)
            # The response structure for batch embedding in google-genai
            # res.embeddings is a list of EmbedContentResponse or similar objects, each has 'values'
            embeddings.extend([e.values for e in res.embeddings])
        return embeddings

    async def embed_query(self, text: str) -> List[float]:
        async def _call_embed():
             response = await self.client.aio.models.embed_content(
                model=self.config.model_name,
                contents=text,
                config=types.EmbedContentConfig(
                     task_type="RETRIEVAL_QUERY",
                     output_dimensionality=self.config.dimensions if getattr(self.config, 'dimensions', None) else None
                )
            )
             return response

        res = await self._retry(_call_embed)
        return res.embeddings[0].values

    async def generate(self, prompt: str, system_instruction: str = "") -> str:
        config_args = {}
        if self.config.temperature is not None:
             config_args['temperature'] = self.config.temperature
        if self.config.max_tokens is not None:
             config_args['max_output_tokens'] = self.config.max_tokens
        
        if system_instruction:
             config_args['system_instruction'] = system_instruction

        async def _call_generate():
            response = await self.client.aio.models.generate_content(
                model=self.config.model_name,
                contents=prompt,
                config=types.GenerateContentConfig(**config_args)
            )
            return response

        response = await self._retry(_call_generate)
        return response.text

    async def generate_stream(self, prompt: str, system_instruction: str = "") -> AsyncGenerator[dict, None]:
        config_args = {}
        if self.config.temperature is not None:
             config_args['temperature'] = self.config.temperature
        if self.config.max_tokens is not None:
             config_args['max_output_tokens'] = self.config.max_tokens
        
        if system_instruction:
             config_args['system_instruction'] = system_instruction

        # google-genai aio generate_content_stream returns an async iterator
        async for chunk in await self.client.aio.models.generate_content_stream(
            model=self.config.model_name,
            contents=prompt,
            config=types.GenerateContentConfig(**config_args)
        ):
            if chunk.text:
                yield { 'delta': chunk.text, 'finish_reason': None, 'usage': None }
